# -*- coding: utf-8 -*-
import yaml
import smuthi.simulation
import numpy as np
import smuthi.index_conversion as idx
import smuthi.coordinates as coord
import os


def read_input_yaml(filename):
    print('\nReading ' + os.path.abspath(filename))
    with open(filename, 'r') as input_file:
        input_data = yaml.load(input_file.read())

    simulation = smuthi.simulation.Simulation()

    simulation.initial_field_collection.vacuum_wavelength = float(input_data['vacuum wavelength'])
    simulation.length_unit = input_data.get('length unit')

    # particle collection
    particle_input = input_data['scattering particles']
    if isinstance(particle_input, str):
        particle_type = 'sphere'
        with open(particle_input, 'r') as particle_specs_file:
            for line in particle_specs_file:
                if len(line.split()) > 0:
                    if line.split()[-1] == 'spheres':
                        particle_type = 'sphere'
                    elif line.split()[-1] == 'spheroids':
                        particle_type = 'spheroid'
                    if not line.split()[0] == '#':
                        numeric_line_data = [float(x) for x in line.split()]
                        pos = numeric_line_data[:3]
                        if particle_type == 'sphere':
                            r = numeric_line_data[3]
                            n = numeric_line_data[4] + 1j * numeric_line_data[5]
                            simulation.particle_collection.add_sphere(radius=r, refractive_index=n, position=pos)
                        if particle_type == 'spheroid':
                            c = numeric_line_data[3]
                            a = numeric_line_data[4]
                            n = numeric_line_data[5] + 1j * numeric_line_data[6]
                            simulation.particle_collection.add_spheroid(semi_axis_c=c, semi_axis_a=a,
                                                                        refractive_index=n, position=pos)
    else:
        for prtcl in input_data['scattering particles']:
            if prtcl['shape'] == 'sphere':
                r = float(prtcl['radius'])
                n = (float(prtcl['refractive index']) + 1j * float(prtcl['extinction coefficient']))
                pos = [float(prtcl['position'][0]), float(prtcl['position'][1]), float(prtcl['position'][2])]
                simulation.particle_collection.add_sphere(radius=r, refractive_index=n, position=pos)
            elif prtcl['shape'] == 'spheroid':
                c = float(prtcl['semi axis c'])
                a = float(prtcl['semi axis a'])
                n = float(prtcl['refractive index']) + 1j * float(prtcl['extinction coefficient'])
                pos = [float(prtcl['position'][0]), float(prtcl['position'][1]), float(prtcl['position'][2])]
                euler_angles = [float(prtcl['euler angles'][0]), float(prtcl['euler angles'][1]),
                                float(prtcl['euler angles'][2])]
                simulation.particle_collection.add_spheroid(semi_axis_c=c, semi_axis_a=a, refractive_index=n,
                                                            position=pos, euler_angles=euler_angles)
            else:
                raise ValueError('Currently, only spheres and spheroids are implemented')

    # layer system
    thick = [float(d) for d in input_data['layer system'][0]['thicknesses']]
    ref_ind = [float(n) for n in input_data['layer system'][0]['refractive indices']]
    ext_coeff = [float(n) for n in input_data['layer system'][0]['extinction coefficients']]
    ref_ind = np.array(ref_ind) + 1j * np.array(ext_coeff)
    ref_ind = ref_ind.tolist()
    simulation.layer_system.__init__(thicknesses=thick, refractive_indices=ref_ind)

    # initial field
    for infld in input_data['initial field']:
        if infld['type'] == 'plane wave':
            a = float(infld['amplitude'])
            if infld['angle units'] == 'degree':
                ang_fac = np.pi / 180
            else:
                ang_fac = 1
            pol_ang = ang_fac * float(infld['polar angle'])
            az_ang = ang_fac * float(infld['azimuthal angle'])
            if infld['polarization'] == 'TE':
                pol = 0
            elif infld['polarization'] == 'TM':
                pol = 1
            else:
                raise ValueError('polarization must be "TE" or "TM"')
            ref = [float(infld['reference point'][0]), float(infld['reference point'][1]),
                   float(infld['reference point'][2])]
            simulation.initial_field_collection.add_planewave(amplitude=a, polar_angle=pol_ang, reference_point=ref,
                                                              azimuthal_angle=az_ang, polarization=pol)

    # linear system
    lmax = int(input_data['lmax'])
    mmax = int(input_data.get('mmax', lmax))
    idx.set_swe_specs(l_max=lmax, m_max=mmax)

    # contour
    neff_waypoints = [complex(nf) for nf in input_data['neff waypoints']]
    neff_discretization = float(input_data['neff discretization'])
    simulation.wr_neff_contour = coord.ComplexContour(neff_waypoints=neff_waypoints,
                                                      neff_discretization=neff_discretization)

    # T-matrix method
    simulation.t_matrix_method = input_data.get('tmatrix method')


    # post processing
    for item in input_data['post processing']:
        if item['task'] == 'evaluate cross sections':
            simulation.post_processing.tasks.append(item)

    return simulation



# -*- coding: utf-8 -*-

import numpy as np
import smuthi.coordinates as coord
import smuthi.layers as lay
import smuthi.vector_wave_functions as vwf
import smuthi.index_conversion as idx


class InitialFieldCollection:
    """Collection of initial field parameter sets."""
    def __init__(self, vacuum_wavelength=None):
        """Initialize

        input:
        vacuum_wavelength   (length unit)
        """
        self.vacuum_wavelength = vacuum_wavelength

        # A list of dictionaries, each corresponding to one element of the initial field. For example, a simulation with
        # N dipole sources would have N entries in the list
        self.specs_list = []

    def add_planewave(self, amplitude=1, polar_angle=None, reference_point=[0, 0, 0], azimuthal_angle=0,
                      polarization=0, k_parallel=0, n_effective=None, layer=None):
        """Add plane wave to the collection of initial field components.

        Redundant entries (like k_parallel AND n_effective) don't need to be specified.

        input:
        amplitude:          complex amplitude of plane wave at x,y,z=0,0,0
        k_parallel:         in-plane wavenumber (inverse length)
        n_effective:        k_parallel = omega * n_effective
        polar_angle:        polar angle in radians
        refractive_index:   refractive index of medium from which plane wave comes
        azimuthal_angle:    azimuthal angle in radians (that is the angle between in-plane k-vector and x-axis)
        polarization:       polarization (0='TE' and E~e_azimuth, 1='TM' and E~e_polar).
        layer:              'top' for top illumination and 'bottom' for bottom illumination
        """
        if polarization == 'TE':
            polarization = 0
        elif polarization == 'TM':
            polarization = 1

        if polar_angle is None:
            if n_effective is None:
                n_effective = k_parallel / coord.angular_frequency(self.vacuum_wavelength)
            if layer == 'top':
                polar_angle = np.pi - np.asin(n_effective)
            elif layer == 'bottom':
                polar_angle = np.asin(n_effective)

        self.specs_list.append({'type': 'plane wave', 'amplitude': amplitude, 'polar angle': polar_angle,
                                'azimuthal angle': azimuthal_angle, 'polarization': polarization,
                                'reference point': reference_point})


def initial_field_swe_coefficients(initial_field_collection, particle_collection, layer_system,
                                   layerresponse_precision=None):
    """Return the spherical wave expansion coefficients a0 for all particles specified in a
    smuthi.particles.ParticleCollection object, for all initial field elements. a0 is a numpy array of dimensions
    NS x nmax, where NS is the number of particles and nmax is the number of swe coefficients per particle."""
    a0 = np.zeros((particle_collection.particle_number(), idx.number_of_indices()), dtype=complex)
    for iS, rS in enumerate(particle_collection.particle_positions()):
        for i0, specs in enumerate(initial_field_collection.specs_list):
            if specs['type'] == 'plane wave':
                a0[iS, :] += planewave_swe_coefficients(vacuum_wavelength=initial_field_collection.vacuum_wavelength,
                                                        amplitude=specs['amplitude'],
                                                        polar_angle=specs['polar angle'],
                                                        azimuthal_angle=specs['azimuthal angle'],
                                                        polarization=specs['polarization'],
                                                        planewave_reference_point=specs['reference point'],
                                                        particle_position=rS, layer_system=layer_system,
                                                        layerresponse_precision=layerresponse_precision)
            else:
                raise ValueError('This initial field type is currently not implemented')

    return a0


def planewave_swe_coefficients(vacuum_wavelength=None, amplitude=1, polar_angle=0, azimuthal_angle=0, polarization=0,
                               planewave_reference_point=[0, 0, 0], particle_position=[0, 0, 0], layer_system=None,
                               layerresponse_precision=None):
    """Return the initial field coefficients (spherical wave expansion) as a numpy-array for a single particle in a
    planarly layered medium and a single initial plane wave. The coefficients array has dimension NS x nmax, where NS is
    the number of particles and nmax is the number of swe coefficients per particle.

    Input:
    vacuum_wavelength           (length unit)
    amplitude                   (default=1)
    polar_angle                 The polar angle also decides if the plane wave comes from the bottom side of the layer
                                system (polar angle<=pi/2) or from the top side (radian, default=0)
    azimuthal_angle             (radian, default=0)
    polarization                (0=TE, 1=TM)
    planewave_reference_point   In the format [x,y,z]. At this point in space the amplitude of the incoming wave is as
                                defined. (length unit, default=[0,0,0])
    particle_position           In the format [x,y,z]. (length unit, default=[0,0,0])
    layer_system                smuthi.layers.LayerSystem object that defines the planarly layered medium in which the
                                particle is located
    layerresponse_precision     If None, standard numpy is used for the layer response. If int>0, that many decimal
                                digits are considered in multiple precision. (default=None)
    """

    l_max = idx.l_max
    m_max = idx.m_max

    angular_frequency = coord.angular_frequency(vacuum_wavelength)
    blocksize = idx.number_of_indices()
    
    #initialize output
    aPR = np.zeros(blocksize, dtype=complex) # layer system mediated
    aPD = np.zeros(blocksize, dtype=complex) # direct
    
    if polar_angle < (np.pi / 2): #then the plane wave comes from the bottom layer 
        iP = 0
    else: #top layer
        iP = layer_system.number_of_layers() - 1
    
    #wavevectors in initial layer iP and in particle layer iS
    k_iP = layer_system.refractive_indices[iP] * angular_frequency
    kp = k_iP * np.sin(polar_angle)
    kx = np.cos(azimuthal_angle) * kp
    ky = np.sin(azimuthal_angle) * kp
    pm_kz_iP = k_iP * np.cos(polar_angle)
    kvec_iP = np.array([kx, ky, pm_kz_iP])

    iS = layer_system.layer_number(particle_position[2])
    k_iS = layer_system.refractive_indices[iS] * angular_frequency
    kz_iS = coord.k_z(k_parallel=kp, k=k_iS)

    kvec_pl_iS = np.array([kx, ky, kz_iS])
    kvec_mn_iS = np.array([kx, ky, -kz_iS])

    rvec_iP = np.array([0, 0, layer_system.reference_z(iP)])
    rvec_iS = np.array([0, 0, layer_system.reference_z(iS)])
    rvec_S = np.array(particle_position)
    rvec_0 = np.array(planewave_reference_point)

    # phase factors
    ejkriP = np.exp(1j * np.dot(kvec_iP, rvec_iP - rvec_0))
    ejkplriSS = np.exp(1j * np.dot(kvec_pl_iS, rvec_S - rvec_iS))
    ejkmnriSS = np.exp(1j * np.dot(kvec_mn_iS, rvec_S - rvec_iS))
    ejkriPS = np.exp(1j * np.dot(kvec_iP, rvec_S))

    L = lay.layersystem_response_matrix(pol=polarization, layer_d=layer_system.thicknesses,
                                        layer_n=layer_system.refractive_indices, kpar=kp, omega=angular_frequency,
                                        fromlayer=iP, tolayer=layer_system.layer_number(particle_position[2]),
                                        precision=layerresponse_precision)
    gR = np.dot(L, np.array([1, 1]))
    for tau in range(2):
        for m in range(-m_max, m_max + 1):
            emjma = np.exp(-1j * m * azimuthal_angle)
            for l in range(max(1, abs(m)), l_max+1):
                n = idx.multi_to_single_index(tau, l, m)
                Bdagpl = vwf.transformation_coefficients_VWF(tau, l, m, pol=polarization, kp=kp, kz=kz_iS, dagger=True)
                Bdagmn = vwf.transformation_coefficients_VWF(tau, l, m, pol=polarization, kp=kp, kz=-kz_iS, dagger=True)
                Bvec = np.array([Bdagpl * ejkplriSS, Bdagmn * ejkmnriSS])
                aPR[n] = 4 * amplitude * emjma * ejkriP * np.dot(Bvec, gR)
                if iS == iP: # add direct contribution
                    Bin = vwf.transformation_coefficients_VWF(tau, l, m, pol=polarization, kp=kp, kz=pm_kz_iP,
                                                              dagger=True)
                    aPD[n] = 4 * amplitude * emjma * ejkriPS * Bin

    return aPR + aPD




# -*- coding: utf-8 -*-
"""Provide routines for multiple scattering."""

import numpy as np
import scipy.special
import smuthi.index_conversion as idx
import smuthi.coordinates as coord
import smuthi.layers as lay
import smuthi.vector_wave_functions as vwf
import smuthi.spherical_functions as sf
import matplotlib.pyplot as plt
import warnings


def layer_mediated_coupling_block(vacuum_wavelength, receiving_particle_position, emitting_particle_position,
                                  layer_system, neff_contour, layerresponse_precision=None, show_integrand=False):
    """Return the layer-system mediated particle coupling matrix W^R for two particles. This routine is explicit, but
    slow.

    Input:
    vacuum_wavelength:              (length unit)
    receiving_particle_position:    In the format [x,y,z] (length unit)
    emitting_particle_position:     In the format [x,y,z] (length unit)
    layer_system:                   An instance of smuthi.layers.LayerSystem describing the stratified medium
    neff_contour:                   An instance of smuthi.coordinates.ComplexContour to define the contour of the
                                    Sommerfeld integral
    layerresponse_precision:        Number of decimal digits (int). If specified, the layer-response is evaluated using
                                    mpmath multiple precision. Otherwise, standard numpy.
    show_integrand:                 If True, the norm of the integrand is plotted.
    """
    omega = coord.angular_frequency(vacuum_wavelength)

    # read out index specs
    lmax = idx.l_max
    mmax = idx.m_max
    blocksize = idx.number_of_indices()

    # cylindrical coordinates of relative position vectors
    rs1 = np.array(receiving_particle_position)
    rs2 = np.array(emitting_particle_position)
    rs2s1 = rs1 - rs2
    rhos2s1 = np.linalg.norm(rs2s1[0:2])
    phis2s1 = np.arctan2(rs2s1[1], rs2s1[0])
    is1 = layer_system.layer_number(rs1[2])
    ziss1 = rs1[2] - layer_system.reference_z(is1)
    is2 = layer_system.layer_number(rs2[2])
    ziss2 = rs2[2] - layer_system.reference_z(is2)

    # wave numbers
    neff = neff_contour.neff()
    kpar = omega * neff
    kis2 = omega * layer_system.refractive_indices[is2]
    kzis1 = coord.k_z(n_effective=neff, vacuum_wavelength=vacuum_wavelength,
                      refractive_index=layer_system.refractive_indices[is1])
    kzis2 = coord.k_z(n_effective=neff, vacuum_wavelength=vacuum_wavelength,
                      refractive_index=layer_system.refractive_indices[is2])

    # phase factors
    ejkz = np.zeros((2, 2, len(neff)), dtype=complex)  # indices are: particle, plus/minus, kpar_idx
    ejkz[0, 0, :] = np.exp(1j * kzis1 * ziss1)
    ejkz[0, 1, :] = np.exp(- 1j * kzis1 * ziss1)
    ejkz[1, 0, :] = np.exp(1j * kzis2 * ziss2)
    ejkz[1, 1, :] = np.exp(- 1j * kzis2 * ziss2)

    # layer response
    L = lay.evaluate_layerresponse_lookup(layer_system.thicknesses, layer_system.refractive_indices, kpar, omega, is2,
                                          is1, layerresponse_precision)  # polarization, pl/mn1, pl/mn2, kpar_idx

    # transformation coefficients
    B = np.zeros((2, 2, 2, blocksize, len(neff)), dtype=complex)  # indices are: particle, pol, plus/minus, n, kpar_idx

    m_vec = np.zeros(blocksize, dtype=int)
    kz_tup = (kzis1, kzis2)
    plmn_tup = (1, -1)
    dagger_tup = (True, False)

    for tau in range(2):
        for m in range(-mmax, mmax + 1):
            for l in range(max(1, abs(m)), lmax + 1):
                n = idx.multi_to_single_index(tau, l, m)
                m_vec[n] = m
                for iprt in range(2):
                    for iplmn, plmn in enumerate(plmn_tup):
                        for pol in range(2):
                            B[iprt, pol, iplmn, n, :] = vwf.transformation_coefficients_VWF(tau, l, m, pol, kpar,
                                                                                            plmn * kz_tup[iprt],
                                                                                            dagger=dagger_tup[iprt])

    BeL = np.zeros((2, 2, blocksize, len(neff)), dtype=complex)  # indices are: pol, plmn2, n1, kpar_idx
    for iplmn1 in range(2):
        for pol in range(2):
            BeL[pol, :, :, :] += (L[pol, iplmn1, :, np.newaxis, :] *
                                     B[0, pol, iplmn1, np.newaxis, :, :] * ejkz[0, iplmn1, :])
    BeLBe = np.zeros((blocksize, blocksize, len(neff)), dtype=complex)  # indices are: n1, n2, kpar_idx
    for iplmn2 in range(2):
        for pol in range(2):
            BeLBe += BeL[pol, iplmn2, :, np.newaxis, :] * B[1, pol, iplmn2, :, :] * ejkz[1, 1 - iplmn2, :]

    # bessel function and jacobi factor
    bessel_list = []
    for dm in range(2 * lmax + 1):
        bessel_list.append(scipy.special.jv(dm, kpar * rhos2s1))
    bessel_full = np.array([[bessel_list[abs(m_vec[n1] - m_vec[n2])]
                             for n1 in range(blocksize)] for n2 in range(blocksize)])
    jacobi_vector = kpar / (kzis2 * kis2)
    integrand = bessel_full * jacobi_vector * BeLBe
    integral = np.trapz(integrand, x=kpar, axis=-1)
    m2_minus_m1 = m_vec - m_vec[np.newaxis].T
    wr = 4 * (1j) ** abs(m2_minus_m1) * np.exp(1j * m2_minus_m1 * phis2s1) * integral

    if show_integrand:
        norm_integrand = np.zeros(len(neff))
        for i in range(len(neff)):
            norm_integrand[i] = 4 * np.linalg.norm(integrand[:, :, i])
        plt.plot(neff.real, norm_integrand)
        plt.show()

    return wr


def layer_mediated_coupling_matrix(vacuum_wavelength, particle_collection, layer_system, neff_contour,
                                   layerresponse_precision=None):
    """Return the layer-system mediated particle coupling matrix W^R for a particle collection.
    This routine is explicit, but slow. It is thus suited for problems with few particles only.

    NOT TESTED

    Input:
    vacuum_wavelength:              (length unit)
    particle_collection:            An instance of  smuthi.particles.ParticleCollection describing the scattering
                                    particles
    layer_system:                   An instance of smuthi.layers.LayerSystem describing the stratified medium
    neff_contour:                   An instance of smuthi.coordinates.ComplexContour to define the contour of the
                                    Sommerfeld integral
    layerresponse_precision:        Number of decimal digits (int). If specified, the layer-response is evaluated using
                                    mpmath multiple precision. Otherwise, standard numpy.
    """
    blocksize = idx.number_of_indices()
    particle_number = particle_collection.particle_number()

    # initialize result
    wr = np.zeros((particle_number, blocksize, particle_number, blocksize), dtype=complex)

    for s1, particle1 in enumerate(particle_collection.particles):
        rs1 = particle1['position']
        for s2, particle2 in enumerate(particle_collection.particles):
            rs2 = particle2['position']
            wrblock = layer_mediated_coupling_block(vacuum_wavelength, rs1, rs2, layer_system, neff_contour,
                                                    layerresponse_precision)

            wr[s1, :, s2, :] = wrblock

    return wr


def direct_coupling_matrix(vacuum_wavelength, particle_collection, layer_system):
    """Return the direct particle coupling matrix W for a particle collection in a layered medium.

    NOT TESTED

    Input:
    vacuum_wavelength:              (length unit)
    particle_collection:            An instance of  smuthi.particles.ParticleCollection describing the scattering
                                    particles
    layer_system:                   An instance of smuthi.layers.LayerSystem describing the stratified medium
    """
    omega = coord.angular_frequency(vacuum_wavelength)

    # indices
    blocksize = idx.number_of_indices()
    particle_number = particle_collection.particle_number()
    lmax = idx.l_max
    mmax = idx.m_max

    # initialize result
    w = np.zeros((particle_number, blocksize, particle_number, blocksize), dtype=complex)

    # check which particles are in same layer
    particle_layer_indices = [[] for x in range(layer_system.number_of_layers())]
    for i_particle, particle in enumerate(particle_collection.particles):
        particle_layer_indices[layer_system.layer_number(particle['position'][2])].append(i_particle)

    # direct coupling inside each layer
    pos_array = np.array(particle_collection.particle_positions())
    for i_layer in range(layer_system.number_of_layers()):
        if len(particle_layer_indices[i_layer]) > 1:

            k = omega * layer_system.refractive_indices[i_layer]

            # coordinates
            x = pos_array[particle_layer_indices[i_layer], 0]
            y = pos_array[particle_layer_indices[i_layer], 1]
            z = pos_array[particle_layer_indices[i_layer], 2]
            dx = x[:, np.newaxis] - x[np.newaxis, :]
            dy = y[:, np.newaxis] - y[np.newaxis, :]
            dz = z[:, np.newaxis] - z[np.newaxis, :]
            d = np.sqrt(dx**2 + dy**2 + dz**2)
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", message="invalid value encountered in true_divide")
                cos_theta = dz / d
                sin_theta = np.sqrt(dx**2 + dy**2) / d
            phi = np.arctan2(dy, dx)

            npart_layer = len(x)
            w_layer = np.zeros((npart_layer, blocksize, npart_layer, blocksize), dtype=complex)
            # indices: receiv. part., receiv. idx, emit. part., emit. idx

            # spherical functions
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", message="Casting complex values to real discards the imaginary part")
                bessel_h = [sf.spherical_hankel(n, k * d) for n in range(2 * lmax + 1)]
            legendre, _, _ = sf.legendre_normalized(cos_theta, sin_theta, 2 * lmax)

            for m1 in range(-mmax, mmax + 1):
                for m2 in range(-mmax, mmax + 1):
                    eimph = np.exp(1j * (m1 - m2) * phi)
                    for l1 in range(max(1, abs(m1)), lmax + 1):
                        for l2 in range(max(1, abs(m2)), lmax + 1):
                            A, B = complex(0), complex(0)
                            for ld in range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1):  # if ld<abs(m1-m2) then P=0
                                a5, b5 = vwf.ab5_coefficients(l1, m1, l2, m2, ld)
                                A += a5 * bessel_h[ld] * legendre[ld][abs(m1 - m2)]
                                B += b5 * bessel_h[ld] * legendre[ld][abs(m1 - m2)]
                            A, B = eimph * A, eimph * B
                            for tau1 in range(2):
                                n1 = idx.multi_to_single_index(tau1, l1, m1)
                                for tau2 in range(2):
                                    n2 = idx.multi_to_single_index(tau2, l2, m2)
                                    if tau1 == tau2:
                                        w_layer[:, n2, :, n1] = A  # remember that w = A.T
                                    else:
                                        w_layer[:, n2, :, n1] = B

            for i1 in range(npart_layer):
                for i2 in range(npart_layer):
                    if not i1 == i2:
                        s1 = particle_layer_indices[i_layer][i1]
                        s2 = particle_layer_indices[i_layer][i2]
                        w[s1, :, s2, :] = w_layer[i1, :, i2, :]

    return w



# -*- coding: utf-8 -*-
"""Provide class to manage a simulation."""

import smuthi.linear_system as lin
import smuthi.particles as part
import smuthi.layers as lay
import smuthi.initial_field as init
import smuthi.t_matrix as tmt
import smuthi.particle_coupling as coup
import smuthi.coordinates as coord
import smuthi.post_processing as pp
import sys
import matplotlib.pyplot as plt
import pkg_resources


class Simulation:
    def __init__(self, layer_system=None, particle_collection=None, initial_field_collection=None, linear_system=None,
                 wr_neff_contour=None, post_processing=None, tmatrix_method=None, length_unit='length unit'):
        """Initialize

        input:
        """
        if layer_system is None:
            layer_system = lay.LayerSystem()
        if particle_collection is None:
            particle_collection = part.ParticleCollection()
        if initial_field_collection is None:
            initial_field_collection = init.InitialFieldCollection()
        if linear_system is None:
            linear_system = lin.LinearSystem()
        if wr_neff_contour is None:
            wr_neff_contour = coord.ComplexContour()
        if post_processing is None:
            post_processing = pp.PostProcessing()
        if tmatrix_method is None:
            tmatrix_method = {}

        self.layer_system = layer_system
        self.particle_collection = particle_collection
        self.initial_field_collection = initial_field_collection
        self.linear_system = linear_system
        self.wr_neff_contour = wr_neff_contour
        self.post_processing = post_processing
        self.tmatrix_method = tmatrix_method
        self.length_unit = length_unit

    def run(self):
        clear_console()
        print(welcome_message())
        self.status_message = welcome_message()

        # compute initial field coefficients
        sys.stdout.write("Compute initial field coefficients ... ")
        sys.stdout.flush()
        self.linear_system.initial_field_coefficients = \
            init.initial_field_swe_coefficients(self.initial_field_collection, self.particle_collection,
                                                self.layer_system)
        sys.stdout.write("done. \n")

        # compute T-matrix
        sys.stdout.write("Compute T-matrices ... ")
        sys.stdout.flush()
        self.linear_system.t_matrices = \
            tmt.t_matrix_collection(self.initial_field_collection.vacuum_wavelength, self.particle_collection,
                                    self.layer_system, self.tmatrix_method)
        sys.stdout.write("done. \n")

        # compute particle coupling matrices
        sys.stdout.write("Compute direct particle coupling matrix ... ")
        sys.stdout.flush()
        self.linear_system.coupling_matrix = coup.direct_coupling_matrix(
            self.initial_field_collection.vacuum_wavelength, self.particle_collection, self.layer_system)
        sys.stdout.write("done. \n")

        sys.stdout.write("Compute layer system mediated particle coupling matrix ... ")
        sys.stdout.flush()
        self.linear_system.coupling_matrix += coup.layer_mediated_coupling_matrix(
            self.initial_field_collection.vacuum_wavelength, self.particle_collection, self.layer_system,
            self.wr_neff_contour)
        sys.stdout.write("done. \n")

        # solve linear system
        sys.stdout.write("Solve linear system ... ")
        sys.stdout.flush()
        self.linear_system.solve()
        sys.stdout.write("done. \n")

        # post processing
        sys.stdout.write("Post processing ... ")
        sys.stdout.flush()
        self.post_processing.run(self)
        sys.stdout.write("done. \n")
        plt.show()


def welcome_message():
    version = pkg_resources.get_distribution("smuthi").version
    msg = ("\n"
           "********************************\n"
           "    SMUTHI version " + version + "\n"
           "********************************\n")
    return msg


def clear_console():
    #print("\n"*100)
    pass


# -*- coding: utf-8 -*-
"""Provide class for the representation of planar layer systems."""

import numpy as np
import sympy

# global variables
matrix_format = np.array
math_module = np
layerresponse_lookup = {'args': {}, 'data': None}


class LayerSystem:
    """Stack of planar layers."""
    def __init__(self, thicknesses=[0, 0], refractive_indices=[1, 1]):
        """Initialize

        input:
        thicknesses         list of layer thicknesses, first and last are semi inf and set to 0 (length unit)
        refractive_indices  list of complex refractive indices in the form n+jk
        """
        self.thicknesses = thicknesses
        self.thicknesses[0] = 0
        self.thicknesses[-1] = 0
        self.refractive_indices = refractive_indices

    def number_of_layers(self):
        """Return total number of layers"""
        return len(self.thicknesses)

    def lower_zlimit(self, i):
        """Return the z-coordinate of lower boundary

        The coordinate system is defined such that z=0 corresponds to the interface between layer 0 and layer 1.

        input:
        i:      index of layer in question (must be between 0 and number_of_layers-1)
        """
        if i == 0:
            return -np.inf
        else:
            sumthick = 0
            for d in self.thicknesses[1:i]:
                sumthick += d
        return sumthick

    def upper_zlimit(self, i):
        """Return the z-coordinate of upper boundary.

        The coordinate system is defined such that z=0 corresponds to the interface between layer 0 and layer 1.

        input:
        i:      index of layer in question (must be between 0 and number_of_layers-1)
        """
        if i == self.number_of_layers() - 1:
            return np.inf
        else:
            sumthick = 0
            for d in self.thicknesses[1:i + 1]:
                sumthick += d
        return sumthick

    def reference_z(self, i):
        """Return the anchor point's z-coordinate.

        The coordinate system is defined such that z=0 corresponds to the interface between layer 0 and layer 1.

        input:
        i:      index of layer in question (must be between 0 and number_of_layers-1)
        """
        if i == 0:
            return self.upper_zlimit(i)
        else:
            return self.lower_zlimit(i)

    def layer_number(self, z):
        """ Return number of layer that contains point [0,0,z]

        If z is on the interface, the higher layer number is selected.

        input:
        z:       z-coordinate of query point (length unit)
        """
        d = 0
        laynum = 0
        for th in self.thicknesses[1:]:
            if z >= d:
                laynum += 1
                d += th
            else:
                break
        return laynum


def fresnel_r(pol, kz1, kz2, n1, n2):
    """Return the complex Fresnel reflection coefficient.

    Input:
    pol     polarization (0=TE, 1=TM)
    kz1     incoming wave's z-wavenumber (k*cos(alpha1))
    kz2     transmitted wave's z-wavenumber (k*cos(alpha2))
    n1      first medium's complex refractive index (n+ik)
    n2      second medium's complex refractive index (n+ik)
    """
    if pol == 0:
        return (kz1 - kz2) / (kz1 + kz2)
    else:
        return (n2 ** 2 * kz1 - n1 ** 2 * kz2) / (n2 ** 2 * kz1 + n1 ** 2 * kz2)


def fresnel_t(pol, kz1, kz2, n1, n2):
    """Return the complex Fresnel transmission coefficient.

    Input:
    pol     polarization (0=TE, 1=TM)
    kz1     incoming wave's z-wavenumber (k*cos(alpha1))
    kz2     transmitted wave's z-wavenumber (k*cos(alpha2))
    n1      first medium's complex refractive index (n+ik)
    n2      second medium's complex refractive index (n+ik)
    """
    if pol == 0:
        return 2 * kz1 / (kz1 + kz2)
    else:
        return 2 * n1 * n2 * kz1 / (n2 ** 2 * kz1 + n1 ** 2 * kz2)


def interface_transition_matrix(pol, kz1, kz2, n1, n2):
    """Return the interface transition matrix to be used in the Transfer matrix algorithm.

    Input:
    pol                 polarization (0=TE, 1=TM)
    kz1                 incoming wave's z-wavenumber (k*cos(alpha1))
    kz2                 transmitted wave's z-wavenumber (k*cos(alpha2))
    n1                  first medium's complex refractive index (n+ik)
    n2                  second medium's complex refractive index (n+ik)
    """
    t = fresnel_t(pol, kz1, kz2, n1, n2)
    r = fresnel_r(pol, kz1, kz2, n1, n2)
    return 1 / t * matrix_format([[1, r], [r, 1]])


def layer_propagation_matrix(kz, d):
    """Return the layer propagation matrix to be used in the Transfer matrix algorithm.

    Input:
    kz                  z-wavenumber (k*cos(alpha))
    d                   thickness of layer
    """
    return matrix_format([[math_module.exp(-1j * kz * d), 0], [0, math_module.exp(1j * kz * d)]])


def layersystem_transfer_matrix(pol, layer_d, layer_n, kpar, omega):
    """Return the transfer matrix of a planarly layered medium.

    Input:
    pol                 polarization(0=TE, 1=TM)
    layer_d             numpy array of layer thicknesses
    layer_n             numpy array of complex layer refractive indices
    kpar                in-plane wavenumber
    omega               angular frequency in units of c=1: omega=2*pi/lambda
    """
    layer_kz = []
    for n in layer_n:
        kz = math_module.sqrt((omega * n) ** 2 - kpar ** 2 + 0j)
        if kz.imag < 0:
            kz = -kz
        layer_kz.append(kz)
    tmat = math_module.eye(2)
    for i in range(len(layer_d) - 1):
        dmat = interface_transition_matrix(pol, layer_kz[i], layer_kz[i + 1], layer_n[i], layer_n[i + 1])
        pmat = layer_propagation_matrix(layer_kz[i], layer_d[i])
        tmat = matrix_product(tmat, matrix_product(pmat, dmat))
    return tmat


def layersystem_scattering_matrix(pol, layer_d, layer_n, kpar, omega):
    """Return the scattering matrix of a planarly layered medium.

    Input:
    pol                 polarization(0=TE, 1=TM)
    layer_d             numpy array of layer thicknesses
    layer_n             numpy array of complex layer refractive indices
    kpar                in-plane wavenumber
    omega               angular frequency in units of c=1: omega=2*pi/lambda
    """
    layer_kz = []
    for n in layer_n:
        kz = math_module.sqrt((omega * n) ** 2 - kpar ** 2 + 0j)
        if kz.imag < 0:
            kz = -kz
        layer_kz.append(kz)
    smat = math_module.eye(2)
    for i in range(len(layer_d) - 1):
        dmat = interface_transition_matrix(pol, layer_kz[i], layer_kz[i + 1], layer_n[i], layer_n[i + 1])
        pmat = layer_propagation_matrix(layer_kz[i], layer_d[i])
        tmat = matrix_product(pmat, dmat)
        s11 = smat[0, 0] / (tmat[0, 0] - smat[0, 1] * tmat[1, 0])
        s12 = (smat[0, 1] * tmat[1, 1] - tmat[0, 1]) / (tmat[0, 0] - smat[0, 1] * tmat[1, 0])
        s21 = smat[1, 1] * tmat[1, 0] * s11 + smat[1, 0]
        s22 = smat[1, 1] * tmat[1, 0] * s12 + smat[1, 1] * tmat[1, 1]
        smat = matrix_format([[s11, s12], [s21, s22]])
    return smat


def layersystem_response_matrix(pol, layer_d, layer_n, kpar, omega, fromlayer, tolayer, precision=None):
    """Return the layer system response matrix of a planarly layered medium.

    If kpar is specified as an array with length N, return N 2x2 matrices as an 2x2xN ndarray.

    Input:
    pol         polarization(0=TE, 1=TM)
    layer_d     list of layer thicknesses
    layer_n     list of complex layer refractive indices
    kpar        (complex float or array-like) in-plane wavenumber
    omega       angular frequency in units of c=1: omega=2*pi/lambda
    fromlayer   number of layer where the excitation is located
    tolayer     number of layer where the response is evaluated
    precision   number of decimal digits to keep in interim calculations
    """
    global matrix_format
    global math_module
    if precision is None:
        matrix_format = np.array
        math_module = np
    else:
        sympy.mpmath.mp.dps = precision
        matrix_format = sympy.mpmath.matrix
        math_module = sympy.mpmath

    if hasattr(kpar, "__len__"):    # is kpar an array? then use recursive call to fill an 2 x 2 x N ndarray
        result = np.zeros((2, 2, len(kpar)), dtype=complex)
        for i, kp in enumerate(kpar):
            result[:, :, i] = layersystem_response_matrix(pol, layer_d, layer_n, kp, omega, fromlayer, tolayer,
                                                          precision)
        return result

    layer_d_above = [0] + layer_d[fromlayer:]
    layer_n_above = [layer_n[fromlayer]] + layer_n[fromlayer:]
    smat_above = layersystem_scattering_matrix(pol, layer_d_above, layer_n_above, kpar, omega)
    layer_d_below = layer_d[: fromlayer] + [0]
    layer_n_below = layer_n[: fromlayer] + [layer_n[fromlayer]]
    smat_below = layersystem_scattering_matrix(pol, layer_d_below, layer_n_below, kpar, omega)
    lmat = matrix_product(matrix_inverse(matrix_format([[1, -smat_below[0, 1]], [-smat_above[1, 0], 1]])),
                          matrix_format([[0, smat_below[0, 1]], [smat_above[1, 0], 0]]))
    if tolayer > fromlayer:
        tmat_fromto = layersystem_transfer_matrix(pol, layer_d[fromlayer:tolayer + 1], layer_n[fromlayer:tolayer + 1],
                                                  kpar, omega)
        lmat = matrix_product(matrix_inverse(tmat_fromto), lmat + matrix_format([[1, 0], [0, 0]]))
    elif tolayer < fromlayer:
        tmat_fromto = layersystem_transfer_matrix(pol, layer_d[tolayer:fromlayer + 1], layer_n[tolayer:fromlayer + 1],
                                                  kpar, omega)
        lmat = matrix_product(tmat_fromto, lmat + matrix_format([[0, 0], [0, 1]]))
    return np.array(lmat.tolist(), dtype=complex)


def matrix_product(m1, m2):
    if isinstance(m1, sympy.mpmath.matrix) and isinstance(m2, sympy.mpmath.matrix):
        return m1 * m2
    elif isinstance(m1, np.ndarray) and isinstance(m2, np.ndarray):
        return np.dot(m1, m2)


def matrix_inverse(m):
    if isinstance(m, sympy.mpmath.matrix):
        return m ** (-1)
    elif isinstance(m, np.ndarray):
        return np.linalg.inv(m)


def evaluate_layerresponse_lookup(layer_d, layer_n, kpar, omega, fromlayer, tolayer, precision):
    """Return the layer system response matrix of a planarly layered medium as a 2x2x2xN ndarray.
    The indices are: polarization, first index of layer response matrix, second index of layer response matrix,
    index of parallel wave vector.
    The result is stored in a lookup table, and returned when the function is called for the second time with the same
    arguments.

    Input:
    layer_d     list of layer thicknesses
    layer_n     list of complex layer refractive indices
    kpar        (complex float or array-like) in-plane wavenumber
    omega       angular frequency in units of c=1: omega=2*pi/lambda
    fromlayer   number of layer where the excitation is located
    tolayer     number of layer where the response is evaluated
    precision   number of decimal digits to keep in interim calculations
    """
    global layerresponse_lookup
    L_args = {'thicknesses': layer_d, 'refractive indices': layer_n, 'kpar': kpar, 'omega': omega,
              'precision': precision}
    laynum = len(layer_d)

    # check for argument equality:
    equal_args = (L_args.get('thicknesses') == layerresponse_lookup['args'].get('thicknesses')
                  and L_args.get('refractive indices') == layerresponse_lookup['args'].get('refractive indices')
                  and np.all(L_args.get('kpar') == layerresponse_lookup['args'].get('kpar'))
                  and L_args.get('omega') == layerresponse_lookup['args'].get('omega')
                  and L_args.get('precision') == layerresponse_lookup['args'].get('precision'))

    if equal_args:
        if layerresponse_lookup['data'][fromlayer][tolayer] is not None:
            L = layerresponse_lookup['data'][fromlayer][tolayer]
        else:
            L = np.zeros((2, 2, 2, len(kpar)), dtype=complex)  # indices are: polarization, pl/mn1, pl/mn2, kpar_idx
            for pol in range(2):
                L[pol, :, :, :] = layersystem_response_matrix(pol, layer_d, layer_n, kpar, omega, fromlayer, tolayer,
                                                              precision)
            layerresponse_lookup['data'][fromlayer][tolayer] = L
    else:
        layerresponse_lookup['args'] = L_args
        layerresponse_lookup['data'] = [[None for x in range(laynum)] for x in range(laynum)]
        L = np.zeros((2, 2, 2, len(kpar)), dtype=complex)  # indices are: polarization, pl/mn1, pl/mn2, kpar_idx
        for pol in range(2):
            L[pol, :, :, :] = layersystem_response_matrix(pol, layer_d, layer_n, kpar, omega, fromlayer, tolayer,
                                                          precision)
        layerresponse_lookup['data'][fromlayer][tolayer] = L

    return L



# -*- coding: utf-8 -*-
import numpy as np
import smuthi.coordinates as coord
import smuthi.index_conversion as idx
import smuthi.vector_wave_functions as vwf
import smuthi.layers as lay
import matplotlib.pyplot as plt


class PostProcessing:
    def __init__(self):
        self.tasks = []

    def run(self, simulation):
        particle_collection = simulation.particle_collection
        linear_system = simulation.linear_system
        layer_system = simulation.layer_system
        initial_field_collection = simulation.initial_field_collection
        for item in self.tasks:
            if item['task'] == 'evaluate cross sections':
                polar_angles = item.get('polar angles')
                azimuthal_angles = item.get('azimuthal angles')
                layerresponse_precision = item.get('layerresponse precision')
                filename_forward = item.get('filename forward')
                filename_backward = item.get('filename backward')

                if (len(initial_field_collection.specs_list) > 1
                    or not initial_field_collection.specs_list[0]['type'] == 'plane wave'):
                    raise ValueError('Cross section only defined for single plane wave excitation.')

                i_top = layer_system.number_of_layers() - 1
                beta_P = initial_field_collection.specs_list[0]['polar angle']
                if beta_P < np.pi / 2:
                    i_P = 0
                    n_P = layer_system.refractive_indices[i_P]
                    n_transm = layer_system.refractive_indices[i_top]
                else:
                    i_P = i_top
                    n_P = layer_system.refractive_indices[i_P]
                    n_transm = layer_system.refractive_indices[0]
                if n_P.imag:
                    raise ValueError('plane wave from absorbing layer: cross section undefined')

                self.scattering_cross_section = scattering_cross_section(polar_angles=polar_angles,
                                                                         initial_field_collection=initial_field_collection,
                                                                         azimuthal_angles=azimuthal_angles,
                                                                         particle_collection=particle_collection,
                                                                         linear_system=linear_system, layer_system=layer_system,
                                                                         layerresponse_precision=layerresponse_precision)

                self.extinction_cross_section = extinction_cross_section(initial_field_collection=initial_field_collection,
                                                                         particle_collection=particle_collection,
                                                                         linear_system=linear_system,
                                                                         layer_system=layer_system,
                                                                         layerresponse_precision=layerresponse_precision)

                # distinguish the cases of top/bottom illumination
                print()
                print('-----------------------------------------------------------------')
                print('Cross sections:')
                if i_P == 0:
                    print('Scattering into bottom layer (diffuse reflection):  ',
                          self.scattering_cross_section['total bottom'][0]
                          + self.scattering_cross_section['total bottom'][1], ' ' + simulation.length_unit + '^2')

                    if n_transm.imag == 0:
                        print('Scattering into top layer (diffuse transmission):   ',
                              self.scattering_cross_section['total top'][0]
                              + self.scattering_cross_section['total top'][1],
                              ' ' + simulation.length_unit + '^2')

                        print('Total scattering cross section:                     ',
                              self.scattering_cross_section['total'][0] + self.scattering_cross_section['total'][1],
                              ' ' + simulation.length_unit + '^2')

                    print('Bottom layer extinction (extinction of reflection): ',
                          self.extinction_cross_section['bottom'].real,
                          ' ' + simulation.length_unit + '^2')

                    if n_transm.imag == 0:
                        print('Top layer extinction (extinction of transmission):  ',
                              self.extinction_cross_section['top'].real,
                              ' ' + simulation.length_unit + '^2')

                        print('Total extinction cross section:                     ',
                              (self.extinction_cross_section['top'] + self.extinction_cross_section['bottom']).real,
                              ' ' + simulation.length_unit + '^2')

                else:
                    print('Scattering into top layer (diffuse reflection):       ',
                          self.scattering_cross_section['total top'][0] + self.scattering_cross_section['total top'][1],
                          ' ' + simulation.length_unit + '^2')

                    if n_transm.imag == 0:
                        print('Scattering into bottom layer (diffuse transmission):  ',
                              self.scattering_cross_section['total bottom'][0] +
                              self.scattering_cross_section['total bottom'][1],
                              ' ' + simulation.length_unit + '^2')

                        print('Total scattering cross section:                       ',
                              self.scattering_cross_section['total'][0] + self.scattering_cross_section['total'][1],
                              ' ' + simulation.length_unit + '^2')

                    print('Top layer extinction (extinction of reflection):      ',
                          self.extinction_cross_section['top'],
                          ' ' + simulation.length_unit + '^2')

                    if n_transm.imag == 0:
                        print('Bottom layer extinction (extinction of transmission): ',
                              self.extinction_cross_section['bottom'],
                              ' ' + simulation.length_unit + '^2')

                        print('Total extinction cross section:                       ',
                              self.extinction_cross_section['top'] + self.extinction_cross_section['bottom'],
                              ' ' + simulation.length_unit + '^2')
                print('-----------------------------------------------------------------')

                if item.get('show plots', False):

                    # dsc as polar plot
                    if layer_system.refractive_indices[i_top].imag == 0:
                        # top layer
                        top_idcs = (self.scattering_cross_section['polar angles'] <= np.pi / 2)
                        alpha_grid, beta_grid = np.meshgrid(self.scattering_cross_section['azimuthal angles'],
                                                            self.scattering_cross_section['polar angles'][top_idcs].real
                                                            * 180 / np.pi)

                        fig = plt.figure()
                        ax = fig.add_subplot(111, polar=True)
                        ax.pcolormesh(alpha_grid, beta_grid, (self.scattering_cross_section['differential'][0, top_idcs, :] +
                                                              self.scattering_cross_section['differential'][1, top_idcs, :]))
                        plt.title('DCS in top layer (' + simulation.length_unit + '^2)')

                        plt.figure()
                        plt.plot(self.scattering_cross_section['polar angles'][top_idcs] * 180 / np.pi,
                                 (self.scattering_cross_section['polar'][0, top_idcs]
                                  + self.scattering_cross_section['polar'][1, top_idcs]) * np.pi / 180)
                        plt.xlabel('polar angle (degree)')
                        plt.ylabel('d_CS/d_beta (' + simulation.length_unit + '^2)')
                        plt.title('Polar differential scattering cross section in top layer')
                        plt.grid(True)

                    if layer_system.refractive_indices[0].imag == 0:
                        # bottom layer
                        bottom_idcs = (self.scattering_cross_section['polar angles'] > np.pi / 2)
                        alpha_grid, beta_grid = np.meshgrid(self.scattering_cross_section['azimuthal angles'],
                                                            self.scattering_cross_section['polar angles'][bottom_idcs].real
                                                            * 180 / np.pi)

                        fig = plt.figure()
                        ax = fig.add_subplot(111, polar=True)
                        ax.pcolormesh(alpha_grid, 180 - beta_grid, (self.scattering_cross_section['differential'][0, bottom_idcs, :] +
                                                                      self.scattering_cross_section['differential'][1, bottom_idcs, :]))
                        plt.title('DCS in bottom layer (' + simulation.length_unit + '^2)')

                        plt.figure()
                        plt.plot(180 - self.scattering_cross_section['polar angles'][bottom_idcs] * 180 / np.pi,
                                 (self.scattering_cross_section['polar'][0, bottom_idcs]
                                  + self.scattering_cross_section['polar'][1, bottom_idcs]) * np.pi / 180)
                        plt.xlabel('polar angle (degree)')
                        plt.ylabel('d_CS/d_beta (' + simulation.length_unit + '^2)')
                        plt.title('Polar differential scattering cross section in bottom layer')
                        plt.grid(True)

                    plt.show()


def scattering_cross_section(polar_angles=None, initial_field_collection=None, azimuthal_angles=None,
                             particle_collection=None, linear_system=None, layer_system=None,
                             layerresponse_precision=None):
    """Evaluate and display the differential scattering cross section as a function of solid angle.
    Return a dictionary scattering_cross_section with the following entries:
    'differential':     Differential cross section as an np.array of dimension 2 x nb x na where nb is the number of
                        polar angles and na is the number of azimuthal angles. The first index refers to polarization.
    'total':            Total cross section as list with two entries (for two polarizations)
    'polar':            Polar differential cross section (that is scattered power per polar angle, divided by incoming
                        intensity)
    'polar angles':     Polar angles for which the differential cross section was evaluated
    'azimuthal angles': Azimuthal angles for which the differential cross section was evaluated
    'forward indices':  The indices of polar angles that correspond to directions in the top hemisphere
    'backward indices': The indices of polar angles that correspond to directions in the bottom hemisphere

    input:
    polar_angles:               (float) array of polar angles (radian), default: from 1 to 180 degree in steps of 1
    initial_field_collection:   smuthi.initial_field.InitialFieldCollection object
    azimuthal_angles:           (float) array of azimuthal angles (radian), default: from 1 to 360 degree in steps of 1
    particle_collection:        smuthi.particles.ParticleCollection object
    linear_system:              smuthi.linear_system.LinearSystem object
    layer_system:               smuthi.layers.LayerSystem object
    layerresponse_precision:    If None, standard numpy is used for the layer response. If int>0, that many decimal
                                digits are considered in multiple precision. (default=None)
    """
    if (len(initial_field_collection.specs_list) > 1
         or not initial_field_collection.specs_list[0]['type'] == 'plane wave'):
        raise ValueError('Cross section only defined for single plane wave excitation.')

    if polar_angles is None:
        polar_angles = (np.concatenate([np.arange(0, 90, 1, dtype=float), np.arange(91, 181, 1, dtype=float)])
                        * np.pi / 180)
    if azimuthal_angles is None:
        azimuthal_angles = np.arange(0, 361, 1, dtype=float) * np.pi / 180

    i_top = layer_system.number_of_layers() - 1
    vacuum_wavelength = initial_field_collection.vacuum_wavelength
    omega = coord.angular_frequency(vacuum_wavelength)
    k_bot = omega * layer_system.refractive_indices[0]
    k_top = omega * layer_system.refractive_indices[-1]

    # read plane wave parameters
    A_P = initial_field_collection.specs_list[0]['amplitude']
    beta_P = initial_field_collection.specs_list[0]['polar angle']
    if beta_P < np.pi / 2:
        i_P = 0
        n_P = layer_system.refractive_indices[i_P]
    else:
        i_P = i_top
        n_P = layer_system.refractive_indices[i_P]
    if n_P.imag:
        raise ValueError('plane wave from absorbing layer: cross section undefined')
    else:
        n_P = n_P.real

    initial_intensity = abs(A_P) ** 2 * abs(np.cos(beta_P)) * n_P / 2

    # differential scattering cross section
    dscs = scattered_far_field(polar_angles, vacuum_wavelength, azimuthal_angles, particle_collection, linear_system,
                               layer_system, layerresponse_precision)['intensity'] / initial_intensity

    top_idcs = polar_angles <= (np.pi / 2)
    bottom_idcs = polar_angles > (np.pi / 2)

    if not k_top.imag == 0:
        dscs[:, top_idcs, :] = 0
    if not k_bot.imag == 0:
        dscs[:, bottom_idcs, :] = 0

    if len(azimuthal_angles) > 2:
        # azimuthal average
        polar_dscs = np.trapz(dscs, azimuthal_angles[None, None, :]) * np.sin(polar_angles[None, :])

        # total scattering cross section
        total_cs_top = np.trapz(polar_dscs[:, top_idcs], polar_angles[None, top_idcs])
        total_cs_bottom = np.trapz(polar_dscs[:, bottom_idcs], polar_angles[None, bottom_idcs])
        total_cs = total_cs_top + total_cs_bottom
    else:
        polar_dscs = None
        total_cs_top = None
        total_cs_bottom = None
        total_cs = None

    scs = {'differential': dscs, 'total': total_cs, 'total top': total_cs_top,
           'total bottom': total_cs_bottom, 'polar': polar_dscs, 'polar angles': polar_angles,
           'azimuthal angles': azimuthal_angles}

    return scs


def extinction_cross_section(initial_field_collection=None, particle_collection=None, linear_system=None,
                             layer_system=None, layerresponse_precision=None):
    """Evaluate and display the differential scattering cross section as a function of solid angle. Return dictionary
    with entries
    'forward':      Extinction in the positinve z-direction (top layer)
    'backward':     Extinction in the negative z-direction (bottom layer)

    polar_angles:               (float) array of polar angles (radian), default: from 1 to 180 degree in steps of 1
    initial_field_collection:   smuthi.initial_field.InitialFieldCollection object
    azimuthal_angles:           (float) array of azimuthal angles (radian), default: from 1 to 360 degree in steps of 1
    particle_collection:        smuthi.particles.ParticleCollection object
    linear_system:              smuthi.linear_system.LinearSystem object
    layer_system:               smuthi.layers.LayerSystem object
    layerresponse_precision:    If None, standard numpy is used for the layer response. If int>0, that many decimal
                                digits are considered in multiple precision. (default=None)
    """
    if (len(initial_field_collection.specs_list) > 1
         or not initial_field_collection.specs_list[0]['type'] == 'plane wave'):
        raise ValueError('Cross section only defined for single plane wave excitation.')

    i_top = layer_system.number_of_layers() - 1
    vacuum_wavelength = initial_field_collection.vacuum_wavelength
    omega = coord.angular_frequency(vacuum_wavelength)
    k_bot = omega * layer_system.refractive_indices[0]
    k_top = omega * layer_system.refractive_indices[-1]

    # read plane wave parameters
    pol_P = initial_field_collection.specs_list[0]['polarization']
    beta_P = initial_field_collection.specs_list[0]['polar angle']
    alpha_P = initial_field_collection.specs_list[0]['azimuthal angle']

    if beta_P < np.pi / 2:
        i_P = 0
        n_P = layer_system.refractive_indices[i_P]
        k_P = k_bot
    else:
        i_P = i_top
        n_P = layer_system.refractive_indices[i_P]
        k_P = k_top
    if n_P.imag:
        raise ValueError('plane wave from absorbing layer: cross section undefined')
    else:
        n_P = n_P.real

    # complex amplitude of initial wave (including phase factor for reference point)
    kappa_P = np.sin(beta_P) * k_P
    kx = np.cos(alpha_P) * kappa_P
    ky = np.sin(alpha_P) * kappa_P
    pm_kz_P = k_P * np.cos(beta_P)
    kvec_P = np.array([kx, ky, pm_kz_P])
    rvec_iP = np.array([0, 0, layer_system.reference_z(i_P)])
    rvec_0 = np.array(initial_field_collection.specs_list[0]['reference point'])
    ejkriP = np.exp(1j * np.dot(kvec_P, rvec_iP - rvec_0))
    A_P = initial_field_collection.specs_list[0]['amplitude'] * ejkriP

    initial_intensity = abs(A_P) ** 2 * abs(np.cos(beta_P)) * n_P / 2

    # bottom extinction
    Lbot = lay.layersystem_response_matrix(pol_P, layer_system.thicknesses, layer_system.refractive_indices, kappa_P,
                                           omega, i_P, 0, layerresponse_precision)
    gRPbot = (Lbot[1, 0] + Lbot[1, 1]) * A_P

    gr_scat_bottom_list = plane_wave_pattern_rs(n_effective=np.array([kappa_P/omega]),
                                                azimuthal_angles=np.array([alpha_P]),
                                                vacuum_wavelength=vacuum_wavelength,
                                                particle_collection=particle_collection,
                                                linear_system=linear_system, layer_system=layer_system,
                                                layer_numbers=[0])
    g_scat_bottom_list = plane_wave_pattern_s(n_effective=np.array([kappa_P/omega]),
                                              azimuthal_angles=np.array([alpha_P]), vacuum_wavelength=vacuum_wavelength,
                                              particle_collection=particle_collection, linear_system=linear_system,
                                              layer_system=layer_system, layer_numbers=[0])
    kz_bot = coord.k_z(k_parallel=kappa_P, k=k_bot)
    g_scat_bottom = gr_scat_bottom_list[0][pol_P, 1, 0, 0] + g_scat_bottom_list[0][pol_P, 1, 0, 0]
    P_bot_ext = 4 * np.pi ** 2 * kz_bot / omega * (gRPbot * np.conj(g_scat_bottom)).real
    bottom_extinction_cs = - P_bot_ext / initial_intensity

    # bottom extinction
    Ltop = lay.layersystem_response_matrix(pol_P, layer_system.thicknesses, layer_system.refractive_indices, kappa_P,
                                           omega, i_P, i_top, layerresponse_precision)
    gRPtop = (Ltop[0, 0] + Ltop[0, 1]) * A_P

    gr_scat_top_list = plane_wave_pattern_rs(n_effective=np.array([kappa_P/omega]),
                                             azimuthal_angles=np.array([alpha_P]), vacuum_wavelength=vacuum_wavelength,
                                             particle_collection=particle_collection, linear_system=linear_system,
                                             layer_system=layer_system, layer_numbers=[i_top])
    g_scat_top_list = plane_wave_pattern_s(n_effective=np.array([kappa_P/omega]), azimuthal_angles=np.array([alpha_P]),
                                           vacuum_wavelength=vacuum_wavelength, particle_collection=particle_collection,
                                           linear_system=linear_system, layer_system=layer_system,
                                           layer_numbers=[i_top])
    kz_top = coord.k_z(k_parallel=kappa_P, k=k_top)
    g_scat_top = gr_scat_top_list[0][pol_P, 0, 0, 0] + g_scat_top_list[0][pol_P, 0, 0, 0]
    P_top_ext = 4 * np.pi ** 2 * kz_top / omega * (gRPtop * np.conj(g_scat_top)).real
    top_extinction_cs = - P_top_ext / initial_intensity

    extinction_cs = {'top': top_extinction_cs, 'bottom': bottom_extinction_cs}

    return extinction_cs


def scattered_far_field(polar_angles=None, vacuum_wavelength=None, azimuthal_angles=None,
                        particle_collection=None, linear_system=None, layer_system=None, layerresponse_precision=None):
    """Return a far_field dictionary with the following entries:
    'intensity':        Radiant far field intensity as ndarray of shape 2 x nb x na where nb =len(polar_angles) and
                        na = len(azimuthal_angles).
                        The indices are:
                        - polarization (0=TE, 1=TM)
                        - polar angle index
                        - azimuthal angle index
    'polar intensity':  Polar far field, that is the power per polar angle as ndarray of shape 2 x nb
    'top power':        Total scattered power into top layer (positive z direction)
    'bottom power':     Total scattered power into bottom layer (negative z direction)
    'polar angles':     Polar angles
    'azimuthal angles': Azimuthal angles

    input:
    polar_angles:           (float) array of polar angles values (radian)
    vacuum_wavelength:      (float, length unit)
    azimuthal_angles:       (float) array of azimuthal angle values (radian)
    particle_collection:    smuthi.particles.ParticleCollection object
    linear_system:          smuthi.linear_system.LinearSystem object
    layer_system:           smuthi.layers.LayerSystem object
    layerresponse_precision:    If None, standard numpy is used for the layer response. If int>0, that many decimal
                                digits are considered in multiple precision. (default=None)
    """
    omega = coord.angular_frequency(vacuum_wavelength)
    if polar_angles is None:
        polar_angles = np.arange(0, 181, 1, dtype=float) * np.pi / 180
    if azimuthal_angles is None:
        azimuthal_angles = np.arange(0, 361, 1, dtype=float) * np.pi / 180

    i_top = layer_system.number_of_layers() - 1
    top_idcs = polar_angles <= (np.pi / 2)
    bottom_idcs = polar_angles > (np.pi / 2)
    neff_top = np.sin(polar_angles[top_idcs]) * layer_system.refractive_indices[i_top]
    neff_bottom = np.sin(polar_angles[bottom_idcs]) * layer_system.refractive_indices[0]

    top_pwp_rs = plane_wave_pattern_rs(n_effective=neff_top, azimuthal_angles=azimuthal_angles,
                                       vacuum_wavelength=vacuum_wavelength, particle_collection=particle_collection,
                                       linear_system=linear_system, layer_system=layer_system, layer_numbers=[i_top],
                                       layerresponse_precision=layerresponse_precision)

    top_pwp_s = plane_wave_pattern_s(n_effective=neff_top, azimuthal_angles=azimuthal_angles,
                                     vacuum_wavelength=vacuum_wavelength, particle_collection=particle_collection,
                                     linear_system=linear_system, layer_system=layer_system, layer_numbers=[i_top])

    bottom_pwp_rs = plane_wave_pattern_rs(n_effective=neff_bottom, azimuthal_angles=azimuthal_angles,
                                          vacuum_wavelength=vacuum_wavelength, particle_collection=particle_collection,
                                          linear_system=linear_system, layer_system=layer_system, layer_numbers=[0],
                                          layerresponse_precision=layerresponse_precision)

    bottom_pwp_s = plane_wave_pattern_s(n_effective=neff_bottom, azimuthal_angles=azimuthal_angles,
                                        vacuum_wavelength=vacuum_wavelength, particle_collection=particle_collection,
                                        linear_system=linear_system, layer_system=layer_system, layer_numbers=[0])

    pwp = np.concatenate([top_pwp_rs[0][:, 0, :, :] + top_pwp_s[0][:, 0, :, :],
                          bottom_pwp_rs[0][:, 1, :, :] + bottom_pwp_s[0][:, 1, :, :]], axis=1)

    k_top = omega * layer_system.refractive_indices[i_top]
    k_0 = omega * layer_system.refractive_indices[0]
    kkz2_top = coord.k_z(n_effective=neff_top, omega=omega, k=k_top) ** 2 * k_top
    kkz2_bottom = coord.k_z(n_effective=neff_bottom, omega=omega, k=k_0) ** 2 * k_0
    kkz2 = np.concatenate([kkz2_top, kkz2_bottom])

    far_field_intensity = (2 * np.pi ** 2 / omega * kkz2[np.newaxis, :, np.newaxis] * abs(pwp) ** 2).real

    # azimuthal average
    if len(azimuthal_angles) > 2:
        polar_far_field = np.trapz(far_field_intensity, azimuthal_angles[None, None, :]) * np.sin(polar_angles[None, :])
        # total scattered power
        total_top_power = np.trapz(polar_far_field[:, top_idcs], polar_angles[None, top_idcs])
        total_bottom_power = np.trapz(polar_far_field[:, bottom_idcs], polar_angles[None, bottom_idcs])
    else:
        polar_far_field = None
        total_top_power = None
        total_bottom_power = None


    far_field = {'intensity': far_field_intensity,
                 'polar intensity': polar_far_field,
                 'top power': total_top_power,
                 'bottom power': total_bottom_power,
                 'polar angles': polar_angles,
                 'azimuthal angles': azimuthal_angles}

    return far_field


def plane_wave_pattern_rs(n_effective=None, azimuthal_angles=None, vacuum_wavelength=None,
                          particle_collection=None, linear_system=None, layer_system=None, layer_numbers=None,
                          layerresponse_precision=None):
    """Layer system response plane wave pattern of the scattered field.
    Return a list of plane wave patterns as ndarrays of shape 2 x 2 x nk x na where nk =len(n_effective) and
    na = len(azimuthal_angles).
    The indices are:
    - polarization (0=TE, 1=TM)
    - up/down (0=upwards propagating +kz, 1=downwards propagation -kz)
    - n_effective index
    - azimuthal angle index
    A plane wave pattern is the integrand of a 2-dimensional expansion of the layer system
    response to the scattered field in PVWFs.

    TODO: TEST

    n_effective:            (float) array of n_eff values, i.e. k_parallel / omega
    azimuthal_angles:       (float) array of azimuthal angle values (radian)
    vacuum_wavelength:      (float, length unit)
    particle_collection:    smuthi.particles.ParticleCollection object
    linear_system:          smuthi.linear_system.LinearSystem object
    layer_system:           smuthi.layers.LayerSystem object
    layer_numbers:          list of layer numbers in which the plane wave pattern shall be computed
    layerresponse_precision:    If None, standard numpy is used for the layer response. If int>0, that many decimal
                                digits are considered in multiple precision. (default=None)
    """
    omega = coord.angular_frequency(vacuum_wavelength)
    neff_grid, azimuthal_angle_grid = np.meshgrid(n_effective, azimuthal_angles, indexing='ij')
    kpar_grid = neff_grid * omega

    # read out index specs
    lmax = idx.l_max
    mmax = idx.m_max
    blocksize = idx.number_of_indices()

    iS_list = [layer_system.layer_number(prtcl['position'][2]) for prtcl in particle_collection.particles]
    iS_unique = list(set(iS_list))

    kpar = n_effective * omega
    kiS = [None for x in range(layer_system.number_of_layers())]
    kziS = [None for x in range(layer_system.number_of_layers())]
    ziS = [None for x in range(layer_system.number_of_layers())]
    # transformation coefficients
    B = [None for x in range(layer_system.number_of_layers())]
    # layer response, indices: from layer, to layer
    L = [[None for x1 in range(layer_system.number_of_layers())] for x2 in range(layer_system.number_of_layers())]

    m_vec = np.zeros(blocksize, dtype=int)
    for i in iS_unique:
        kiS[i] = omega * layer_system.refractive_indices[i]
        kziS[i] = coord.k_z(k_parallel=kpar, k=kiS[i])
        ziS[i] = layer_system.reference_z(i)
        # transformation coefficients
        B[i] = np.zeros((2, 2, blocksize, len(n_effective)), dtype=complex)  # indices: pol, plus/minus, n, kpar_idx

        for tau in range(2):
            for m in range(-mmax, mmax + 1):
                for l in range(max(1, abs(m)), lmax + 1):
                    n = idx.multi_to_single_index(tau, l, m)
                    m_vec[n] = m
                    for pol in range(2):
                        B[i][pol, 0, n, :] = vwf.transformation_coefficients_VWF(tau, l, m, pol, kpar, kziS[i])
                        B[i][pol, 1, n, :] = vwf.transformation_coefficients_VWF(tau, l, m, pol, kpar, -kziS[i])

        for i_to in layer_numbers:
            L[i][i_to] = np.zeros((2, 2, 2, len(n_effective)), dtype=complex)  # idcs: pol, pl/mn1, pl/mn2, kpar_idx
            for pol in range(2):
                    L[i][i_to][pol, :, :, :] = lay.layersystem_response_matrix(pol, layer_system.thicknesses,
                                                                               layer_system.refractive_indices, kpar,
                                                                               omega, i, i_to, layerresponse_precision)

    pwp = np.zeros((2, 2, len(n_effective), len(azimuthal_angles)), dtype=complex)  # pol, pl/mn, kp_idx, alpha_idx
    pwp_list = [pwp for x in range(len(layer_numbers))]
    eima = np.exp(1j * np.tensordot(m_vec, azimuthal_angles, axes=0))  # indices: n, alpha_idx

    for iprt, prtcl in enumerate(particle_collection.particles):
        i = layer_system.layer_number(prtcl['position'][2])
        pos = prtcl['position']
        emnikrs_grid = np.exp(-1j * kpar_grid * (pos[0] * np.cos(azimuthal_angle_grid)  # indices: kpar_idx, alpha_idx
                                                 + pos[1] * np.sin(azimuthal_angle_grid)))
        ziSS = pos[2] - ziS[i]
        emnikz = np.exp(-1j * kziS[i] * ziSS)                       # indices: kpar_idx
        eikz = np.exp(1j * kziS[i] * ziSS)                          # indices: kpar_idx

        fac0 = 1 / (2 * np.pi) * emnikrs_grid                        # indices: kpar_idx, alpha_idx
        fac1 = 1 / (kziS[i] * kiS[i])                                # index: kpar_idx
        b = linear_system.scattered_field_coefficients[np.newaxis, iprt, :, np.newaxis]      # index: 1,n, 1
        beB = np.zeros((2, 2, blocksize, len(n_effective)), dtype=complex)  # indices: pol, plus/minus, n, kpar_idx
        beB[:, 0, :, :] = b * (emnikz * B[i][:, 0, :, :])
        beB[:, 1, :, :] = b * (eikz * B[i][:, 1, :, :])                 # idcs: pol, pl/mn, n, kpar_idx
        eimabeB = np.tensordot(beB, eima, axes=[2, 0])               # idcs: pol, pl/mn, kpar_idx, alpha_idx

        for i_ln, i_to in enumerate(layer_numbers):
            for plmn1 in range(2):
                for plmn2 in range(2):
                    pwp_list[i_ln][:, plmn1, :, :] += (fac0[np.newaxis, :, :] *
                                                       (fac1[np.newaxis, :, np.newaxis] *
                                                        (L[i][i_to][:, plmn1, plmn2, :, np.newaxis] *
                                                         eimabeB[:, plmn2, :, :])))

    return pwp_list


def plane_wave_pattern_s(n_effective=None, azimuthal_angles=None, vacuum_wavelength=None, particle_collection=None,
                         linear_system=None, layer_system=None, layer_numbers=None):
    """Plane wave pattern of the direct scattered field.
    Return a list of plane wave patterns as ndarrays of shape 2 x 2 x nk x na where nk =len(n_effective) and
    na = len(azimuthal_angles).
    The indices are:
    - polarization (0=TE, 1=TM)
    - up/down (0=upwards propagating +kz, 1=downwards propagation -kz)
    - n_effective index
    - azimuthal angle index
    A plane wave pattern is the integrand of a 2-dimensional expansion of the direct
    scattered field in PVWFs. In each layer, the plane wave pattern is only valid below all particles (minus-component)
    or above all particles (plus-component)

    TODO: CHECK IF VALID!!


    n_effective:            (float) array of n_eff values, i.e. k_parallel / omega
    azimuthal_angles:       (float) array of azimuthal angle values (radian)
    vacuum_wavelength:      (float, length unit)
    particle_collection:    smuthi.particles.ParticleCollection object
    linear_system:          smuthi.linear_system.LinearSystem object
    layer_system:           smuthi.layers.LayerSystem object
    layer_numbers:          list of layer numbers in which the plane wave pattern shall be computed
    layerresponse_precision:    If None, standard numpy is used for the layer response. If int>0, that many decimal
                                digits are considered in multiple precision. (default=None)
    """
    omega = coord.angular_frequency(vacuum_wavelength)
    neff_grid, azimuthal_angle_grid = np.meshgrid(n_effective, azimuthal_angles, indexing='ij')
    kpar_grid = neff_grid * omega

    # read out index specs
    lmax = idx.l_max
    mmax = idx.m_max
    blocksize = idx.number_of_indices()

    kpar = n_effective * omega

    # plane wave pattern
    pwp = np.zeros((2, 2, len(n_effective), len(azimuthal_angles)), dtype=complex)  # pol, pl/mn, kp_idx, alpha_idx
    pwp_list = [pwp for x in range(len(layer_numbers))]

    m_vec = np.zeros(blocksize, dtype=int)
    for tau in range(2):
        for m in range(-mmax, mmax + 1):
            for l in range(max(1, abs(m)), lmax + 1):
                n = idx.multi_to_single_index(tau, l, m)
                m_vec[n] = m
    eima = np.exp(1j * np.tensordot(m_vec, azimuthal_angles, axes=0))  # indices: n, alpha_idx

    for iidx, i in enumerate(layer_numbers):
        kiS = omega * layer_system.refractive_indices[i]
        kziS = coord.k_z(k_parallel=kpar, k=kiS)
        ziS = layer_system.reference_z(i)
        # transformation coefficients
        B = np.zeros((2, 2, blocksize, len(n_effective)), dtype=complex)  # indices are: pol, plus/minus, n, kpar_idx

        for tau in range(2):
            for m in range(-mmax, mmax + 1):
                for l in range(max(1, abs(m)), lmax + 1):
                    n = idx.multi_to_single_index(tau, l, m)
                    for pol in range(2):
                        B[pol, 0, n, :] = vwf.transformation_coefficients_VWF(tau, l, m, pol, kpar, kziS)
                        B[pol, 1, n, :] = vwf.transformation_coefficients_VWF(tau, l, m, pol, kpar, -kziS)

        for iprt, prtcl in enumerate(particle_collection.particles):
            if layer_system.layer_number(prtcl['position'][2]) == i:
                pos = prtcl['position']
                kx_grid = kpar_grid * np.cos(azimuthal_angle_grid)
                ky_grid = kpar_grid * np.sin(azimuthal_angle_grid)
                emnikplrs_grid = np.exp(-1j * (kx_grid * pos[0] + ky_grid * pos[1]
                                               + kziS[:, np.newaxis] * (pos[2] - ziS)))
                emnikmnrs_grid = np.exp(-1j * (kx_grid * pos[0] + ky_grid * pos[1]
                                               - kziS[:, np.newaxis] * (pos[2] - ziS)))
                eirks = np.concatenate([emnikplrs_grid[np.newaxis, np.newaxis, :, :],
                                        emnikmnrs_grid[np.newaxis, np.newaxis, :, :]], axis=1)  # idcs: kp, al

                fac1 = 1 / (2 * np.pi * kziS * kiS)                          # index: kpar_idx

                b = linear_system.scattered_field_coefficients[iprt, :]     # index: n

                bB = b[np.newaxis, np.newaxis, :, np.newaxis] * B       # idcs: pol, pl/mn, n, kp

                pwp = fac1[np.newaxis, np.newaxis, :, np.newaxis] * eirks * np.tensordot(bB, eima, axes=[2, 0])
                pwp_list[iidx] += pwp

    return pwp_list



# -*- coding: utf-8 -*-

import numpy as np
import scipy.special


def legendre_normalized(ct, st, lmax):
    """Return the normalized associated Legendre function P_l^m(cos \theta) and the angular functions
    \pi_l^m(cos \theta) and \tau_l^m(cos \theta), as defined in
    A. Doicu, T. Wriedt, and Y. A. Eremin: "Light Scattering by Systems of Particles", Springer-Verlag, 2006.
    Two arguments (ct and st) are passed such that the function is valid for general complex arguments, while the branch
    cuts are defined by the user already in the definition of st.

    Input:
    ct      Array: cosine of theta (or kz/k)
    st      Array: sine of theta (or kp/k), need to have same dimension as ct, and st**2+ct**2=1 is assumed
    lmax    Integer: maximal multipole order

    Output:
    plm     List: plm[l][m] contains P_l^m(cos \theta). The entries of the list have same dimension as ct (and st)
    pilm    List: pilm[l][m] contains \pi_l^m(cos \theta).
    taulm   List: taulm[l][m] contains \tau_l^m(cos \theta).
    """
    zr = ct - ct
    plm = [[zr for m in range(lmax + 1)] for l in range(lmax + 1)]
    pilm = [[zr for m in range(lmax + 1)] for l in range(lmax + 1)]
    taulm = [[zr for m in range(lmax + 1)] for l in range(lmax + 1)]
    pprimel0 = [zr for l in range(lmax + 1)]

    plm[0][0] += np.sqrt(2)/2
    plm[1][0] = np.sqrt(3/2) * ct
    pprimel0[1] = np.sqrt(3) * plm[0][0]
    taulm[0][0] = -st * pprimel0[0]
    taulm[1][0] = -st * pprimel0[1]

    for l in range(1, lmax):
        plm[l + 1][0] = (1 / (l + 1) * np.sqrt((2 * l + 1) * (2 * l + 3)) * ct * plm[l][0] -
                         l / (l + 1) * np.sqrt((2 * l + 3) / (2 * l - 1)) * plm[l-1][0])
        pprimel0[l + 1] = ((l + 1) * np.sqrt((2 * (l + 1) + 1) / (2 * (l + 1) - 1)) * plm[l][0] +
                           np.sqrt((2 * (l + 1) + 1) / (2 * (l + 1) - 1)) * ct * pprimel0[l])
        taulm[l + 1][0] = -st * pprimel0[l + 1]

    for m in range(1, lmax + 1):
        plm[m][m] = np.sqrt((2 * m + 1) / (2 * factorial(2 * m))) * double_factorial(2 * m - 1) * st**m
        pilm[m][m] = np.sqrt((2 * m + 1) / (2 * factorial(2 * m))) * double_factorial(2 * m - 1) * st**(m - 1)
        taulm[m][m] = m * ct * pilm[m][m]
        for l in range(m, lmax):
            plm[l + 1][m] = (np.sqrt((2 * l + 1) * (2 * l + 3) / ((l + 1 - m) * (l + 1 + m))) * ct * plm[l][m] -
                             np.sqrt((2 * l + 3) * (l - m) * (l + m) / ((2 * l - 1) * (l + 1 - m) * (l + 1 + m))) *
                             plm[l - 1][m])
            pilm[l + 1][m] = (np.sqrt((2 * l + 1) * (2 * l + 3) / (l + 1 - m) / (l + 1 + m)) * ct * pilm[l][m] -
                              np.sqrt((2 * l + 3) * (l - m) * (l + m) / (2 * l - 1) / (l + 1 - m) / (l + 1 + m)) *
                              pilm[l - 1][m])
            taulm[l + 1][m] = ((l + 1) * ct * pilm[l + 1][m] -
                               (l + 1 + m) * np.sqrt((2 * (l + 1) + 1) * (l + 1 - m) / (2 * (l + 1) - 1) / (l + 1 + m))
                               * pilm[l][m])

    return plm, pilm, taulm


def spherical_bessel(n, x):
    """Return the spherical Bessel function.

    As soon as some bug for complex arguments is resolved, this can be replaced by scipy.special.spherical_jn.
    https://github.com/ContinuumIO/anaconda-issues/issues/1415

    Input:
    n   Int: Order of spherical Bessel function
    x   Complex or float array: Argument for Bessel function
    """
    sphj = scipy.special.sph_jn
    if hasattr(x, "__iter__"):
        j_n = np.array([spherical_bessel(n, v) for v in x])
    else:
        j_n = sphj(n, x)[0][n]
    return j_n


def spherical_hankel(n, x):
    """Return the spherical Hankel function of first kind.

    Input:
    n   Int: Order of spherical Hankel function
    x   Complex or float array: Argument for spherical Hankel function
    """
    sphj = scipy.special.sph_jn
    sphy = scipy.special.sph_yn
    if hasattr(x, "__iter__"):
        h_n = np.array([spherical_hankel(n, v) for v in x])
    else:
        h_n = sphj(n, x)[0][n] + 1j * sphy(n, x)[0][n]
    return h_n


def dx_xj(n, x):
    """Return the derivative of x*j_n(x), where j_n(x) is the spherical Bessel function.

    Input:
    n   Int (n>0): Order of spherical Bessel function
    x   Complex or float array: Argument for spherical Hankel function
    """
    res = x * spherical_bessel(n - 1, x) - n * spherical_bessel(n, x)
    return res


def dx_xh(n, x):
    """Return the derivative of x*h_n(x), where h_n(x) is the spherical Hankel function of first kind.

    Input:
    n   Int (n>0): Order of spherical Bessel function
    x   Complex or float array: Argument for spherical Hankel function
    """
    res = x * spherical_hankel(n - 1, x) - n * spherical_hankel(n, x)
    return res


def factorial(n):
    """Return factorial."""
    if n == 0:
        return 1
    else:
        return n * factorial(n-1)


def double_factorial(n):
    """Return double factorial."""
    if n in (0, 1):
        return 1
    else:
        return n * double_factorial(n - 2)



# -*- coding: utf-8 -*-
"""Functions to map the multiple coefficient indices tau,l,m to a single index n and vice versa."""

# the global variable index_order specifies the order of the multipole indices (tau, l, m) of the spherical wave
# expansion
# index_order = 'tlm' implies the following map between multi-index and single-index:
# n | tau, l, m
# -------------
# 1 | 1, 1, -1
# 2 | 1, 1, 0
# 3 | 1, 1, 1
# 2 | 1, 2, -2
# 2 | 1, 2, -1
# 2 | 1, 2, 0
# ..| ... ...
# ..| 1, lmax, lmax
# ..| 2, 1, -1
# ..| ... ...
index_order = 'tlm'

# global variable for maximal multipole degree
l_max = None

# global variable for maximal multipole order
m_max = None


def multi_to_single_index(tau, l, m):
    """Return a unique single index for the totality of indices characterizing a svwf expansion coefficient.

    input:
    tau:                SVWF polarization (0=spherical TE, 1=spherical TM)
    l:                  SVWF degree (1, ..., lmax)
    m:                  SVWF order (-l,...,l)
    """
    if index_order == 'tlm':
        # use:
        # \sum_{l=1}^lmax (2\min(l,mmax)+1) = \sum_{l=1}^mmax (2l+1) + \sum_{l=mmax+1}^lmax (2mmax+1)
        #                                   = 2*(1/2*mmax*(mmax+1))+mmax  +  (lmax-mmax)*(2*mmax+1)
        #                                   = mmax*(mmax+2)               +  (lmax-mmax)*(2*mmax+1)

        tau_blocksize = m_max * (m_max + 2) + (l_max - m_max) * (2 * m_max + 1)
        n = tau * tau_blocksize
        if (l - 1) <= m_max:
            n += (l - 1) * (l - 1 + 2)
        else:
            n += m_max * (m_max + 2) + (l - 1 - m_max) * (2 * m_max + 1)
        n += m + min(l, m_max)
        return n


def number_of_indices():
    """Return the total number of indices which is the maximal index plus 1."""
    return multi_to_single_index(tau=1, l=l_max, m=m_max) + 1


def set_swe_specs(**kwargs):
    """
    Set the truncation degree and order of the spherical wave expansion, as well as the index order for the single to
    multi index mapping.

    key-word input:
    l_max:              set the global truncation degree of SVWF expansions
    m_max:              set the global truncation order of SVWF expansions. None means m_max = l_max
    index_order:        set string to globally specify the order according to which the indices are arranged
                        Possible choices are:
                        'tlm' (default), which stands for 1. tau, 2. l, 3. m
                        (Other choices are not implemented at the moment.)
    """
    global l_max, m_max, index_order

    # set global variables
    if 'l_max' in kwargs:
        l_max = kwargs['l_max']
        m_max = l_max
    if 'm_max' in kwargs:
        m_max = kwargs['m_max']
    if 'index_order' in kwargs:
        index_order = kwargs['index_order']
    for key in kwargs.keys():
        if not (key == 'l_max' or key == 'm_max' or key == 'index_order'):
            raise ValueError('Unknown argument ' + key)



# -*- coding: utf-8 -*-

import numpy as np
import sympy.physics.wigner
import sympy
import smuthi.spherical_functions as sf


def plane_vector_wave_function(x, y, z, kp, alpha, kz, pol):
    """Return the electric field components of plane wave (PVWF).

    The input arrays should have one of the following dimensions
    x,y,z: (N x 1) matrix
    kp,alpha,kz: (1 x M) matrix
    Ex, Ey, Ez: (M x N) matrix

    or
    x,y,z: (M x N) matrix
    kp,alpha,kz: scalar
    Ex, Ey, Ez: (M x N) matrix

    Input:
    x       numpy-array: x-coordinate of position where to test the field (length unit)
    y       numpy-array: y-coordinate of position where to test the field
    z       numpy-array: z-coordinate of position where to test the field
    kp      numpy-array: parallel component of k-vector (inverse length unit)
    alpha   numpy-array: azimthal angle of k-vector (rad)
    kz      numpy-array: z-component of k-vector (inverse length unit)
    pol     integer: (0=TE, 1=TM)

    Output:
    Ex      numpy-array: x-coordinate of PVWF electric field
    Ey      numpy-array: y-coordinate of PVWF electric field
    Ez      numpy-array: z-coordinate of PVWF electric field
    """
    k = np.sqrt(kp**2 + kz**2)
    kx = kp * np.cos(alpha)
    ky = kp * np.sin(alpha)

    scalar_wave = np.exp(1j * (kx * x + ky * y + kz * z))

    if pol == 0:
        Ex = -np.sin(alpha) * scalar_wave
        Ey = np.cos(alpha) * scalar_wave
        Ez = scalar_wave - scalar_wave
    elif pol == 1:
        Ex = np.cos(alpha) * kz / k * scalar_wave
        Ey = np.sin(alpha) * kz / k * scalar_wave
        Ez = -kp / k * scalar_wave
    else:
        raise ValueError('Polarization must be 0 (TE) or 1 (TM)')

    return Ex, Ey, Ez


def spherical_vector_wave_function(x, y, z, k, nu, tau, l, m):
    """Return the electric field components of spherical vector wave function (SVWF). The conventions are chosen
    according to
    A. Doicu, T. Wriedt, and Y. A. Eremin: "Light Scattering by Systems of Particles", Springer-Verlag, 2006.

    Input:
    x       numpy-array: x-coordinate of position where to test the field (length unit)
    y       numpy-array: y-coordinate of position where to test the field
    z       numpy-array: z-coordinate of position where to test the field
    k       scalar: wavenumber (inverse length unit)
    nu      integer: 1 for regular waves, 3 for outgoing waves
    tau     integer: spherical polarization, 0 for spherical TE and 1 for spherical TM
    l       integer: l=1,... multipole degree (polar quantum number)
    m       integer: m=-l,...,l multipole order (azimuthal quantum number)

    Output:
    Ex      numpy-array: x-coordinate of SVWF electric field
    Ey      numpy-array: y-coordinate of SVWF electric field
    Ez      numpy-array: z-coordinate of SVWF electric field
    """
    r = np.sqrt(x**2 + y**2 + z**2)
    theta = np.arccos(z / r)
    phi = np.arctan2(y, x)

    # unit vector in r-direction
    er_x = x / r
    er_y = y / r
    er_z = z / r

    # unit vector in theta-direction
    eth_x = np.cos(theta) * np.cos(phi)
    eth_y = np.cos(theta) * np.sin(phi)
    eth_z = -np.sin(theta)

    # unit vector in phi-direction
    eph_x = -np.sin(phi)
    eph_y = np.cos(phi)
    eph_z = x - x

    cos_thet = np.cos(theta)
    sin_thet = np.sin(theta)
    plm_list, pilm_list, taulm_list = sf.legendre_normalized(cos_thet, sin_thet, l)
    plm = plm_list[l][abs(m)]
    pilm = pilm_list[l][abs(m)]
    taulm = taulm_list[l][abs(m)]

    kr = k * r
    if nu == 1:
        bes = sf.spherical_bessel(l, kr)
        dxxz = sf.dx_xj(l, kr)
    elif nu == 3:
        bes = sf.spherical_hankel(l, kr)
        dxxz = sf.dx_xh(l, kr)
    else:
        raise ValueError('nu must be 1 (regular SVWF) or 3 (outgoing SVWF)')

    eimphi = np.exp(1j * m * phi)
    prefac = 1/np.sqrt(2 * l * (l + 1))
    if tau == 0:
        Ex = prefac * bes * (1j * m * pilm * eth_x - taulm * eph_x) * eimphi
        Ey = prefac * bes * (1j * m * pilm * eth_y - taulm * eph_y) * eimphi
        Ez = prefac * bes * (1j * m * pilm * eth_z - taulm * eph_z) * eimphi
    elif tau == 1:
        Ex = prefac * (l * (l + 1) * bes / kr * plm * er_x +
                       dxxz / kr * (taulm * eth_x + 1j * m * pilm * eph_x)) * eimphi
        Ey = prefac * (l * (l + 1) * bes / kr * plm * er_y +
                       dxxz / kr * (taulm * eth_y + 1j * m * pilm * eph_y)) * eimphi
        Ez = prefac * (l * (l + 1) * bes / kr * plm * er_z +
                       dxxz / kr * (taulm * eth_z + 1j * m * pilm * eph_z)) * eimphi
    else:
        raise ValueError('tau must be 0 (spherical TE) or 1 (spherical TM)')

    return Ex, Ey, Ez


def transformation_coefficients_VWF(tau, l, m, pol, kp=None, kz=None, pilm_list=None, taulm_list=None, dagger=False):
    """Return the transformation coefficients B to expand SVWF in PVWF and vice versa. See theory part of documentation.

    Input:
    tau         integer: SVWF polarization, 0 for spherical TE, 1 for spherical TM
    l           integaer: l=1,... SVWF multipole degree
    m           integaer: m=-l,...,l SVWF multipole order
    pol         integer: PVWF polarization, 0 for TE, 1 for TM
    kp          complex numpy-array: PVWF in-plane wavenumbers
    kz          complex numpy-array: PVWF out-of-plane wavenumbers
    pilm_list   2D list numpy-arrays: alternatively to kp and kz, pilm and taulm as generated with legendre_normalized
                can directly be handed
    taulm_list  2D list numpy-arrays: alternatively to kp and kz, pilm and taulm as generated with legendre_normalized
                can directly be handed
    dagger      logical, switch on when expanding PVWF in SVWF and off when expanding SVWF in PVWF
    """
    if pilm_list is None:
        k = np.sqrt(kp**2 + kz**2)
        ct = kz / k
        st = kp / k
        plm_list, pilm_list, taulm_list = sf.legendre_normalized(ct, st, l)

    if tau == pol:
        sphfun = taulm_list[l][abs(m)]
    else:
        sphfun = m * pilm_list[l][abs(m)]

    if dagger:
        if pol == 0:
            prefac = -1 / (-1j) ** (l + 1) / np.sqrt(2 * l * (l + 1)) * (-1j)
        elif pol ==1:
            prefac = -1 / (-1j) ** (l + 1) / np.sqrt(2 * l * (l + 1)) * 1
        else:
            raise ValueError('pol must be 0 (TE) or 1 (TM)')
    else:
        if pol == 0:
            prefac = -1 / (1j) ** (l + 1) / np.sqrt(2 * l * (l + 1)) * (1j)
        elif pol ==1:
            prefac = -1 / (1j) ** (l + 1) / np.sqrt(2 * l * (l + 1)) * 1
        else:
            raise ValueError('pol must be 0 (TE) or 1 (TM)')

    B = prefac * sphfun

    return B


def translation_coefficients_svwf(l1, m1, l2, m2, k, d, sph_hankel=None, legendre=None, exp_immphi=None):
    """Return the coefficients of the translation operator for the expansion of an outgoing spherical wave in therms of
    regular spherical waves with respect to a different origin.
    The output is a tuple (A,B), where the translation operator
        trans = \delta_pp' * A + (1-\delta_pp') * B

    Input:
    l1          integaer: l=1,...: Original wave's SVWF multipole degree
    m1          integaer: m=-l,...,l: Original wave's SVWF multipole order
    l2          integaer: l=1,...: Partial wave's SVWF multipole degree
    m2          integaer: m=-l,...,l: Partial wave's SVWF multipole order
    k           complex: wavenumber (inverse length unit)
    d           translation vectors in format [dx, dy, dz] (length unit)
                dx, dy, dz can be scalars or ndarrays
    sph_hankel  list: sph_hankel[i] contains the spherical hankel funciton of degree i, evaluated at k*d where d is the
                norm of the distance vector(s)
    legendre    list of lists: legendre[l][m] contains the legendre function of order l and degree m, evaluated at
                cos(theta) where theta is the polar angle(s) of the distance vector(s)
    """
    # spherical coordinates of d:
    dd = np.sqrt(d[0] ** 2 + d[1] ** 2 + d[2] ** 2)

    if exp_immphi is None:
        phid = np.arctan2(d[1], d[0])
        eimph = np.exp(1j * (m1 - m2) * phid)
    else:
        eimph = exp_immphi[m1][m2]

    if sph_hankel is None:
        sph_hankel = [sf.spherical_hankel(n, k * dd) for n in range(l1 + l2 + 1)]

    if legendre is None:
        costthetd = d[2] / dd
        sinthetd = np.sqrt(d[0] ** 2 + d[1] ** 2) / dd
        legendre, _, _ = sf.legendre_normalized(costthetd, sinthetd, l1 + l2)

    A, B = complex(0), complex(0)
    for ld in range(abs(l1 - l2), l1 + l2 + 1):
        a5, b5 = ab5_coefficients(l1, m1, l2, m2, ld)
        A += a5 * sph_hankel[ld] * legendre[ld][abs(m1 - m2)]
        B += b5 * sph_hankel[ld] * legendre[ld][abs(m1 - m2)]
    A, B = eimph * A, eimph * B
    return A, B


def ab5_coefficients(l1, m1, l2, m2, p, symbolic=False):
    """Return a tuple (a5, b5) where a5 and b5 are the coefficients used in the evaluation of the SVWF translation
    operator. The computation is based on the sympy.physics.wigner package and is performed with symbolic numbers.
    If symbolic=True is specified as input argument, symbolic numbers are returned. Otherwise, complex (default).
    """
    jfac = sympy.I ** (abs(m1 - m2) - abs(m1) - abs(m2) + l2 - l1 + p) * (-1) ** (m1 - m2)
    fac1 = sympy.sqrt((2 * l1 + 1) * (2 * l2 + 1) / sympy.S(2 * l1 * (l1 + 1) * l2 * (l2 + 1)))
    fac2a = (l1 * (l1 + 1) + l2 * (l2 + 1) - p * (p + 1)) * sympy.sqrt(2 * p + 1)
    fac2b = sympy.sqrt((l1 + l2 + 1 + p) * (l1 + l2 + 1 - p) * (p + l1 - l2) * (p - l1 + l2) * (2 * p + 1))
    wig1 = sympy.physics.wigner.wigner_3j(l1, l2, p, m1, -m2, -(m1 - m2))
    wig2a = sympy.physics.wigner.wigner_3j(l1, l2, p, 0, 0, 0)
    wig2b = sympy.physics.wigner.wigner_3j(l1, l2, p - 1, 0, 0, 0)

    if symbolic:
        a = jfac * fac1 * fac2a * wig1 * wig2a
        b = jfac * fac1 * fac2b * wig1 * wig2b
    else:
        a = complex(jfac * fac1 * fac2a * wig1 * wig2a)
        b = complex(jfac * fac1 * fac2b * wig1 * wig2b)
    return a, b





# -*- coding: utf-8 -*-
"""Provide class and routines for the solution of the master equation."""

import numpy as np
import scipy.linalg


class LinearSystem:
    """Linear equation for the scattered field swe coefficients."""
    def __init__(self):

        # numpy.ndarray of dimension (NS, nmax)
        self.scattered_field_coefficients = None

        # numpy.ndarray of dimension (NS, nmax)
        self.initial_field_coefficients = None

        # numpy.ndarray of dimension (NS, nmax, nmax)
        # indices are: particle number, outgoing swe index, regular swe index
        self.t_matrices = None

        # numpy.ndarray of dimension (NS, nmax, NS, nmax)
        # indices are: receiving particle number, regular swe index, emitting particle number, outgoing swe index
        self.coupling_matrix = None

        # how to solve?
        self.solver = 'LU'

        # only relevant if solver='LU'
        self.LU_piv = None

    def solve(self):
        """Compute scattered field coefficients"""
        if self.solver == 'LU':
            if self.LU_piv is None:
                lu, piv = scipy.linalg.lu_factor(self.master_matrix(), overwrite_a=False)
                self.LU_piv = (lu, piv)
            bvec = scipy.linalg.lu_solve(self.LU_piv, self.right_hand_side())
            self.scattered_field_coefficients = bvec.reshape(self.initial_field_coefficients.shape)
        else:
            raise ValueError('This solver type is currently not implemented.')

    def right_hand_side(self):
        tai = self.t_matrices * self.initial_field_coefficients[:, np.newaxis, :]
        tai = tai.sum(axis=2)
        return np.concatenate(tai)

    def master_matrix(self):
        NS = len(self.t_matrices[:, 0, 0])
        blocksize = len(self.t_matrices[0, 0, :])
        mm = np.eye(NS * blocksize, dtype=complex)
        w = np.reshape(self.coupling_matrix, (NS * blocksize, NS * blocksize))
        for s in range(NS):
            t = self.t_matrices[s, :, :]
            wblockrow = w[s * blocksize:((s + 1) * blocksize), :]
            twbl = np.dot(t, wblockrow)
            mm[s * blocksize:((s + 1) * blocksize), :] -= twbl
        return mm



# -*- coding: utf-8 -*-
import numpy as np


class ComplexContour:
    """Trajectory of n_effective in the complex plane for the evaluation of Sommerfeld integrals."""
    def __init__(self, neff_waypoints=[0, 1], neff_discretization=1e-2):
        # n_effective waypoints, that is, points through which the contour goes (linear between them)
        # k_parallel = n_effective * omega
        self.neff_waypoints = neff_waypoints

        # Discretization between the waypoints. Either as an array or as a scalar (uniform for all segments)
        self.neff_discretization = neff_discretization

    def neff(self):
        """Return numpy-array of n_effective values that define the contour."""
        neff_segments = []
        for i in range(len(self.neff_waypoints)-1):
            if hasattr(self.neff_discretization, "__len__"):
                npoints = abs(self.neff_waypoints[i+1] - self.neff_waypoints[i]) / self.neff_discretization[i] + 1
            else:
                npoints = abs(self.neff_waypoints[i + 1] - self.neff_waypoints[i]) / self.neff_discretization + 1
            neff_segments.append(self.neff_waypoints[i] + np.linspace(0, 1, num=npoints, endpoint=True, dtype=complex) *
                                 (self.neff_waypoints[i + 1] - self.neff_waypoints[i]))
        return np.concatenate(neff_segments)


def k_z(k_parallel=None, n_effective=None, k=None, omega=None, vacuum_wavelength=None, refractive_index=None):
    """Return z-component of wavevector with k_z.imag >= 0

    Input:
    k_parallel:         in-plane wavenumber (inverse length)
    n_effective:        k_parallel = n_effective * omega
    k:                  wavenumber (inverse length)
    omega:              angular frequency (inverse length, c=1)
    vacuum_wavelength:  (length)
    refractive_index:   refractive index of material in which k_z is evaluated
    """
    if k_parallel is None:
        if omega is None:
            omega = angular_frequency(vacuum_wavelength)
        k_parallel = n_effective * omega

    if k is None:
        if omega is None:
            omega = angular_frequency(vacuum_wavelength)
        k = refractive_index * omega

    kz = np.sqrt(k ** 2 - k_parallel ** 2 + 0j)
    kz = (kz.imag >= 0) * kz + (kz.imag < 0) * (-kz)  # Branch cut such to prohibit negative imaginary
    return kz


def angular_frequency(vacuum_wavelength):
    """Angular frequency in the units of c=1 (time=length)"""
    return 2 * np.pi / vacuum_wavelength


# -*- coding: utf-8 -*-

import numpy as np
import smuthi.spherical_functions
import smuthi.index_conversion
import smuthi.nfmds.t_matrix_axsym as nftaxs


def mie_coefficient(tau, l, k_medium, k_particle, radius):
    """Return the Mie coefficients of a sphere.

    Input:
    tau         integer: spherical polarization, 0 for spherical TE and 1 for spherical TM
    l           integer: l=1,... multipole degree (polar quantum number)
    k_medium    float or complex: wavenumber in surrounding medium (inverse length unit)
    k_particle  float or complex: wavenumber inside sphere (inverse length unit)
    radius      float: radius of sphere (length unit)
    """
    jlkr_medium = smuthi.spherical_functions.spherical_bessel(l, k_medium * radius)
    jlkr_particle = smuthi.spherical_functions.spherical_bessel(l, k_particle * radius)
    dxxj_medium = smuthi.spherical_functions.dx_xj(l, k_medium * radius)
    dxxj_particle = smuthi.spherical_functions.dx_xj(l, k_particle * radius)

    hlkr_medium = smuthi.spherical_functions.spherical_hankel(l, k_medium * radius)
    dxxh_medium = smuthi.spherical_functions.dx_xh(l, k_medium * radius)

    if tau == 0:
        q = (jlkr_medium * dxxj_particle - jlkr_particle * dxxj_medium) / (jlkr_particle * dxxh_medium - hlkr_medium *
                                                                           dxxj_particle)
    elif tau == 1:
        q = ((k_medium ** 2 * jlkr_medium * dxxj_particle - k_particle ** 2 * jlkr_particle * dxxj_medium) /
             (k_particle ** 2 * jlkr_particle * dxxh_medium - k_medium ** 2 * hlkr_medium * dxxj_particle))
    else:
        raise ValueError('tau must be 0 (spherical TE) or 1 (spherical TM)')

    return q


def t_matrix_sphere(k_medium, k_particle, radius):
    """Return the T-matrix of a spherical scattering object.

    Input:
    k_medium            float or complex: wavenumber in surrounding medium (inverse length unit)
    k_particle          float or complex: wavenumber inside sphere (inverse length unit)
    radius              float: radius of sphere (length unit)
    """
    lmax = smuthi.index_conversion.l_max
    mmax = smuthi.index_conversion.m_max
    blocksize = smuthi.index_conversion.number_of_indices()
    t = np.zeros((blocksize, blocksize), dtype=complex)
    for tau in range(2):
        for m in range(-mmax, mmax + 1):
            for l in range(max(1, abs(m)), lmax+1):
                n = smuthi.index_conversion.multi_to_single_index(tau, l, m)
                t[n, n] = mie_coefficient(tau, l, k_medium, k_particle, radius)

    return t


def t_matrix(vacuum_wavelength, n_medium, particle, method={}):
    """Return the T-matrix of a particle.

    NOT TESTED

    Input:
    vacuum_wavelength:  float (length units)
    n_medium:           float or complex: refractive index of surrounding medium
    particle:           Dictionary containing the particle parameters. See smuthi.particles for details.
    method:             Dictionary containing the parameters for the algorithm in case of non-spherical particles.
    """
    if particle['shape'] == 'sphere':
        k_medium = 2 * np.pi / vacuum_wavelength * n_medium
        k_particle = 2 * np.pi / vacuum_wavelength * particle['refractive index']
        radius = particle['radius']
        t = t_matrix_sphere(k_medium, k_particle, radius)
    elif particle['shape'] == 'spheroid':
        if not particle['euler angles'] == [0, 0, 0]:
            raise ValueError('T-matrix for rotated particles currently not implemented.')
        t = nftaxs.tmatrix_spheroid(vacuum_wavelength=vacuum_wavelength, layer_refractive_index=n_medium,
                                    particle_refractive_index=particle['refractive index'],
                                    semi_axis_c=particle['semi axis c'], semi_axis_a=particle['semi axis a'],
                                    use_ds=method.get('use discrete sources', True), nint=method.get('nint', 200),
                                    nrank=method.get('nrank', smuthi.index_conversion.l_max + 2))
    else:
        raise ValueError('T-matrix for ' + particle['shape'] + ' currently not implemented.')

    return t


def rotate_t_matrix(t, euler_angles):
    """Placeholder for a proper T-matrix rotation routine"""
    if euler_angles == [0, 0, 0]:
        return t
    else:
        raise ValueError('Non-trivial rotation not yet implemented')


def t_matrix_collection(vacuum_wavelength, particle_collection, layer_system, method=None):
    """Return the T-matrices for all particles as a numpy.ndarray, in the format (NS, blocksize, blocksize) where NS is
    the number of particles and blocksize is the number of SWE terms per particle.

    Input:
    vacuum_wavelength:      (length unit)
    particle_collection:    An instance of  smuthi.particles.ParticleCollection describing the scattering particles
    layer_system:           An instance of smuthi.layers.LayerSystem describing the stratified medium
    method:                 A dictionary specifying the numerical method to compute T-matrix (in case of non-spherical
                            particles)
    """
    blocksize = smuthi.index_conversion.number_of_indices()
    particle_number = particle_collection.particle_number()
    t_matrices = np.zeros((particle_number, blocksize, blocksize), dtype=complex)

    particle_params_table = []
    tmatrix_table = []

    for ip, particle in enumerate(particle_collection.particles):
        # gather relevant parameters and omit those that don't alter the T-matrix (before rotation)
        # in the first place, just omit position and euler angle, but add refractive index of surrounding medium
        if particle['shape'] == 'sphere':
            params = ['sphere', particle['radius'], particle['refractive index']]
        elif particle['shape'] == 'spheroid':
            params = ['spheroid', particle['semi axis c'], particle['semi axis a'], particle['refractive index']]
        else:
            raise ValueError('invalid particle type: so far only spheres and spheroids are implemented')
        zS = particle['position'][2]
        iS = layer_system.layer_number(zS)
        n_medium = layer_system.refractive_indices[iS]
        params.append(n_medium)

        # compare parameters to table
        for i_old, old_params in enumerate(particle_params_table):
            if params == old_params:
                tmatrix = tmatrix_table[i_old]
                break
        else:  # compute T-matrix and update tables
            tmatrix = t_matrix(vacuum_wavelength, n_medium, particle, method)
            particle_params_table.append(params)
            tmatrix_table.append(tmatrix)

        tmatrix_rotated = rotate_t_matrix(t=tmatrix, euler_angles=particle['euler angles'])
        t_matrices[ip, :, :] = tmatrix_rotated

    return t_matrices



# -*- coding: utf-8 -*-
"""Provide class for the representation of scattering particles."""


class ParticleCollection:
    """Collection of scattering particles."""
    def __init__(self):
        """ A list of dictionaries that contain the following entries:
        - 'shape':              'sphere', 'spheroid', 'finite cylinder'
        - 'refractive index'    Complex refractive index in the form n+kj
        - 'position':           In the form [x, y, z] (length unit)
        - 'euler angles':       Euler angles of rotated particle in the format [alpha, beta, gamma] (radian)
        - further shape-specific parameters which characterize the geometry (like radius etc., see adding methods)
        """
        self.particles = []

    def add_sphere(self, radius, refractive_index, position):
        """Add a sphere to the collection.

        input parameters:
        radius:             sphere radius (length unit)
        refractive_index:   complex refractive index in the form n+jk
        position:           center position in format [x,y,z] (length unit)
        """
        self.particles.append({'shape': 'sphere', 'radius': radius, 'refractive index': refractive_index,
                               'position': position, 'euler angles': [0, 0, 0]})

    def add_spheroid(self, semi_axis_c, semi_axis_a, refractive_index, position, euler_angles=[0, 0, 0]):
        """Add a spheroid to the collection.

        input parameters:
        semi_axis_c:        spheroid semi axis along symmetry axis (length unit)
        semi_axis_a:        spheroid semi axis a=b in transverse direction (length unit)
        refractive_index:   complex refractive index in the form n+jk
        position:           center position in format [x,y,z] (length unit)
        euler_angles:       todo, default=[0,0,0]
        """
        self.particles.append({'shape': 'spheroid', 'semi axis c': semi_axis_c, 'semi axis a': semi_axis_a,
                               'refractive index': refractive_index, 'position': position,
                               'euler angles': euler_angles})

    def add_finite_cylinder(self, cylinder_radius, cylinder_height, refractive_index, position, euler_angles=[0, 0, 0]):
        """Add finite cylinder to the collection.

        input parameters:
        cylinder_radius:    cylinder radius (length unit)
        cylinder_height:    height of cylinder (length unit)
        refractive_index:   complex refractive index in the form n+jk
        position:           center position in format [x,y,z] (length unit)
        euler_angles:       todo, default=[0,0,0]
        """
        self.particles.append({'shape': 'finite cylinder', 'cylinder radius': cylinder_radius,
                               'cylinder height': cylinder_height, 'refractive index': refractive_index,
                               'position': position, 'euler angles': euler_angles})

    def remove_particle(self, i):
        """Remove i-th particle from collection"""
        del self.particles[i]

    def particle_number(self):
        """Return total number of particles in collection"""
        return len(self.particles)

    def particle_positions(self):
        """Return a list of particle positions"""
        return [self.particles[i]['position'] for i in range(self.particle_number())]
